@@ -10,7 +10,6 @@ import GHC.Driver.Env
 import GHC.Driver.Session
 import GHC.Driver.Config.Core.Lint
 import GHC.Driver.Config.Core.Opt.Arity
-import GHC.Tc.Utils.Env
 import GHC.Types.Var
 import GHC.Utils.Outputable ( alwaysQualify )
 
@@ -19,14 +18,9 @@ import GHC.CoreToStg.Prep
 initCorePrepConfig :: HscEnv -> IO CorePrepConfig
 initCorePrepConfig hsc_env = do
    let dflags = hsc_dflags hsc_env
-   convertNumLit <- do
-     let platform = targetPlatform dflags
-         home_unit = hsc_home_unit hsc_env
-         lookup_global = lookupGlobal hsc_env
-     mkConvertNumLiteral platform home_unit lookup_global
    return $ CorePrepConfig
       { cp_catchNonexhaustiveCases = gopt Opt_CatchNonexhaustiveCases dflags
-      , cp_convertNumLit = convertNumLit
+      , cp_platform  = targetPlatform dflags
       , cp_arityOpts = if gopt Opt_DoCleverArgEtaExpansion dflags
                        then Just (initArityOpts dflags)
                        else Nothing
