@@ -1,5 +1,5 @@
 -- Character classification
-{-# LANGUAGE CPP #-}
+
 module GHC.Parser.CharClass
         ( is_ident      -- Char# -> Bool
         , is_symbol     -- Char# -> Bool
@@ -14,8 +14,6 @@ module GHC.Parser.CharClass
         , hexDigit, octDecDigit
         ) where
 
-#include "HsVersions.h"
-
 import GHC.Prelude
 
 import Data.Char        ( ord, chr )
@@ -38,7 +36,7 @@ cDigit  = 64
 
 {-# INLINABLE is_ctype #-}
 is_ctype :: Word8 -> Char -> Bool
-is_ctype mask c = (charType c .&. mask) /= 0
+is_ctype mask c = c <= '\127' && (charType c .&. mask) /= 0
 
 is_ident, is_symbol, is_any, is_space, is_lower, is_upper, is_digit,
     is_alphanum :: Char -> Bool
