@@ -6,10 +6,12 @@ module GHC.Runtime.Context
    , extendInteractiveContextWithIds
    , setInteractivePrintName
    , substInteractiveContext
+   , replaceImportEnv
+   , icReaderEnv
    , icExtendGblRdrEnv
    , icInteractiveModule
    , icInScopeTTs
-   , icPrintUnqual
+   , icNamePprCtx
    )
 where
 
@@ -17,21 +19,20 @@ import GHC.Prelude
 
 import GHC.Hs
 
-import GHC.Driver.Session
+import GHC.Driver.DynFlags
 import {-# SOURCE #-} GHC.Driver.Plugins
 
-import GHC.Runtime.Eval.Types ( Resume )
+import GHC.Runtime.Eval.Types ( IcGlobalRdrEnv(..), Resume )
 
 import GHC.Unit
 import GHC.Unit.Env
 
 import GHC.Core.FamInstEnv
-import GHC.Core.InstEnv ( ClsInst, identicalClsInstHead )
+import GHC.Core.InstEnv
 import GHC.Core.Type
 
-import GHC.Types.Avail
+import GHC.Types.DefaultEnv ( DefaultEnv, emptyDefaultEnv )
 import GHC.Types.Fixity.Env
-import GHC.Types.Id ( isRecordSelector )
 import GHC.Types.Id.Info ( IdDetails(..) )
 import GHC.Types.Name
 import GHC.Types.Name.Env
@@ -43,7 +44,6 @@ import GHC.Types.Var
 import GHC.Builtin.Names ( ioTyConName, printName, mkInteractiveModule )
 
 import GHC.Utils.Outputable
-import GHC.Utils.Misc
 
 {-
 Note [The interactive package]
@@ -67,7 +67,7 @@ This scheme deals well with shadowing.  For example:
 Here we must display info about constructor A, but its type T has been
 shadowed by the second declaration.  But it has a respectable
 qualified name (Ghci1.T), and its source location says where it was
-defined.
+defined, and it can also be used with the qualified name.
 
 So the main invariant continues to hold, that in any session an
 original name M.T only refers to one unique thing.  (In a previous
@@ -95,7 +95,7 @@ The details are a bit tricky though:
    call to initTc in initTcInteractive, which in turn get the module
    from it 'icInteractiveModule' field of the interactive context.
 
-   The 'homeUnitId' field stays as 'main' (or whatever -this-unit-id says.
+   The 'homeUnitId' field stays as 'main' (or whatever -this-unit-id says).
 
  * The main trickiness is that the type environment (tcg_type_env) and
    fixity envt (tcg_fix_env), now contain entities from all the
@@ -182,6 +182,39 @@ e.g.    Prelude> data T = A | B
         Prelude> instance Eq T where ...   -- This one overrides
 
 It's exactly the same for type-family instances.  See #7102
+
+Note [icReaderEnv recalculation]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+The GlobalRdrEnv describing what’s in scope at the prompts consists
+of all the imported things, followed by all the things defined on the prompt,
+with shadowing. Defining new things on the prompt is easy: we shadow as needed,
+and then extend the environment.
+
+But changing the set of imports, which can happen later as well, is tricky
+we need to re-apply the shadowing from all the things defined at the prompt!
+
+For example:
+
+    ghci> let empty = True
+    ghci> import Data.IntMap.Strict     -- Exports 'empty'
+    ghci> empty   -- Still gets the 'empty' defined at the prompt
+    True
+
+It would be correct to re-construct the env from scratch based on
+`ic_tythings`, but that'd be quite expensive if there are many entries in
+`ic_tythings` that shadow each other.
+
+Therefore we keep around a `GlobalRdrEnv` in `igre_prompt_env` that contains
+_just_ the things defined at the prompt, and use that in `replaceImportEnv` to
+rebuild the full env.  Conveniently, `shadowNames` takes such an `OccEnv`
+to denote the set of names to shadow.
+
+INVARIANT: Every `OccName` in `igre_prompt_env` is present unqualified as well
+(else it would not be right to pass `igre_prompt_env` to `shadowNames`.)
+
+The definition of the IcGlobalRdrEnv type should conceptually be in this module,
+and made abstract, but it’s used in `Resume`, so it lives in GHC.Runtime.Eval.Type.
+
 -}
 
 -- | Interactive context, recording information about the state of the
@@ -200,7 +233,7 @@ data InteractiveContext
              -- See Note [The interactive package]
 
          ic_imports :: [InteractiveImport],
-             -- ^ The GHCi top-level scope (ic_rn_gbl_env) is extended with
+             -- ^ The GHCi top-level scope (icReaderEnv) is extended with
              -- these imports
              --
              -- This field is only stored here so that the client
@@ -210,16 +243,23 @@ data InteractiveContext
 
          ic_tythings   :: [TyThing],
              -- ^ TyThings defined by the user, in reverse order of
-             -- definition (ie most recent at the front)
+             -- definition (ie most recent at the front).
+             -- Also used in GHC.Tc.Module.runTcInteractive to fill the type
+             -- checker environment.
              -- See Note [ic_tythings]
 
-         ic_rn_gbl_env :: GlobalRdrEnv,
-             -- ^ The cached 'GlobalRdrEnv', built by
-             -- 'GHC.Runtime.Eval.setContext' and updated regularly
-             -- It contains everything in scope at the command line,
-             -- including everything in ic_tythings
+         ic_gre_cache :: IcGlobalRdrEnv,
+             -- ^ Essentially the cached 'GlobalRdrEnv'.
+             --
+             -- The GlobalRdrEnv contains everything in scope at the command
+             -- line, both imported and everything in ic_tythings, with the
+             -- correct shadowing.
+             --
+             -- The IcGlobalRdrEnv contains extra data to allow efficient
+             -- recalculation when the set of imports change.
+             -- See Note [icReaderEnv recalculation]
 
-         ic_instances  :: ([ClsInst], [FamInst]),
+         ic_instances  :: (InstEnv, [FamInst]),
              -- ^ All instances and family instances created during
              -- this session.  These are grabbed en masse after each
              -- update to be sure that proper overlapping is retained.
@@ -230,10 +270,10 @@ data InteractiveContext
          ic_fix_env :: FixityEnv,
             -- ^ Fixities declared in let statements
 
-         ic_default :: Maybe [Type],
-             -- ^ The current default types, set by a 'default' declaration
+         ic_default :: DefaultEnv,
+             -- ^ The current default classes and types, set by 'default' declarations
 
-          ic_resume :: [Resume],
+         ic_resume :: [Resume],
              -- ^ The stack of breakpoint contexts
 
          ic_monad      :: Name,
@@ -246,9 +286,9 @@ data InteractiveContext
          ic_cwd :: Maybe FilePath,
              -- ^ virtual CWD of the program
 
-         ic_plugins :: ![LoadedPlugin]
+         ic_plugins :: !Plugins
              -- ^ Cache of loaded plugins. We store them here to avoid having to
-             -- load them everytime we switch to the interctive context.
+             -- load them every time we switch to the interactive context.
     }
 
 data InteractiveImport
@@ -261,6 +301,11 @@ data InteractiveImport
       -- of this module, including the things imported
       -- into it.
 
+emptyIcGlobalRdrEnv :: IcGlobalRdrEnv
+emptyIcGlobalRdrEnv = IcGlobalRdrEnv
+    { igre_env = emptyGlobalRdrEnv
+    , igre_prompt_env = emptyGlobalRdrEnv
+    }
 
 -- | Constructs an empty InteractiveContext.
 emptyInteractiveContext :: DynFlags -> InteractiveContext
@@ -268,52 +313,66 @@ emptyInteractiveContext dflags
   = InteractiveContext {
        ic_dflags     = dflags,
        ic_imports    = [],
-       ic_rn_gbl_env = emptyGlobalRdrEnv,
+       ic_gre_cache  = emptyIcGlobalRdrEnv,
        ic_mod_index  = 1,
        ic_tythings   = [],
-       ic_instances  = ([],[]),
+       ic_instances  = (emptyInstEnv,[]),
        ic_fix_env    = emptyNameEnv,
        ic_monad      = ioTyConName,  -- IO monad by default
        ic_int_print  = printName,    -- System.IO.print by default
-       ic_default    = Nothing,
+       ic_default    = emptyDefaultEnv,
        ic_resume     = [],
        ic_cwd        = Nothing,
-       ic_plugins    = []
+       ic_plugins    = emptyPlugins
        }
 
+icReaderEnv :: InteractiveContext -> GlobalRdrEnv
+icReaderEnv = igre_env . ic_gre_cache
+
 icInteractiveModule :: InteractiveContext -> Module
 icInteractiveModule (InteractiveContext { ic_mod_index = index })
-  = mkInteractiveModule index
+  = mkInteractiveModule (show index)
 
 -- | This function returns the list of visible TyThings (useful for
--- e.g. showBindings)
+-- e.g. showBindings).
+--
+-- It picks only those TyThings that are not shadowed by later definitions on the interpreter,
+-- to not clutter :showBindings with shadowed ids, which would show up as Ghci9.foo.
+--
+-- Some TyThings define many names; we include them if _any_ name is still
+-- available unqualified.
 icInScopeTTs :: InteractiveContext -> [TyThing]
-icInScopeTTs = ic_tythings
+icInScopeTTs ictxt = filter in_scope_unqualified (ic_tythings ictxt)
+  where
+    in_scope_unqualified thing = or
+        [ unQualOK gre
+        | gre <- tyThingLocalGREs thing
+        , let name = greName gre
+        , Just gre <- [lookupGRE_Name (icReaderEnv ictxt) name]
+        ]
 
--- | Get the PrintUnqualified function based on the flags and this InteractiveContext
-icPrintUnqual :: UnitEnv -> InteractiveContext -> PrintUnqualified
-icPrintUnqual unit_env InteractiveContext{ ic_rn_gbl_env = grenv } =
-    mkPrintUnqualified unit_env grenv
+-- | Get the NamePprCtx function based on the flags and this InteractiveContext
+icNamePprCtx :: UnitEnv -> InteractiveContext -> NamePprCtx
+icNamePprCtx unit_env ictxt = mkNamePprCtx ptc unit_env (icReaderEnv ictxt)
+  where ptc = initPromotionTickContext (ic_dflags ictxt)
 
 -- | extendInteractiveContext is called with new TyThings recently defined to update the
--- InteractiveContext to include them.  Ids are easily removed when shadowed,
--- but Classes and TyCons are not.  Some work could be done to determine
--- whether they are entirely shadowed, but as you could still have references
--- to them (e.g. instances for classes or values of the type for TyCons), it's
--- not clear whether removing them is even the appropriate behavior.
+-- InteractiveContext to include them. By putting new things first, unqualified
+-- use will pick the most recently defined thing with a given name, while
+-- still keeping the old names in scope in their qualified form (Ghci1.foo).
 extendInteractiveContext :: InteractiveContext
                          -> [TyThing]
-                         -> [ClsInst] -> [FamInst]
-                         -> Maybe [Type]
+                         -> InstEnv -> [FamInst]
+                         -> DefaultEnv
                          -> FixityEnv
                          -> InteractiveContext
 extendInteractiveContext ictxt new_tythings new_cls_insts new_fam_insts defaults fix_env
   = ictxt { ic_mod_index  = ic_mod_index ictxt + 1
                             -- Always bump this; even instances should create
                             -- a new mod_index (#9426)
-          , ic_tythings   = new_tythings ++ old_tythings
-          , ic_rn_gbl_env = ic_rn_gbl_env ictxt `icExtendGblRdrEnv` new_tythings
-          , ic_instances  = ( new_cls_insts ++ old_cls_insts
+          , ic_tythings   = new_tythings ++ ic_tythings ictxt
+          , ic_gre_cache  = ic_gre_cache ictxt `icExtendIcGblRdrEnv` new_tythings
+          , ic_instances  = ( new_cls_insts `unionInstEnv` old_cls_insts
                             , new_fam_insts ++ fam_insts )
                             -- we don't shadow old family instances (#7102),
                             -- so don't need to remove them here
@@ -321,42 +380,48 @@ extendInteractiveContext ictxt new_tythings new_cls_insts new_fam_insts defaults
           , ic_fix_env    = fix_env  -- See Note [Fixity declarations in GHCi]
           }
   where
-    new_ids = [id | AnId id <- new_tythings]
-    old_tythings = filterOut (shadowed_by new_ids) (ic_tythings ictxt)
-
     -- Discard old instances that have been fully overridden
     -- See Note [Override identical instances in GHCi]
     (cls_insts, fam_insts) = ic_instances ictxt
-    old_cls_insts = filterOut (\i -> any (identicalClsInstHead i) new_cls_insts) cls_insts
+    old_cls_insts = filterInstEnv (\i -> not $ anyInstEnv (identicalClsInstHead i) new_cls_insts) cls_insts
 
 extendInteractiveContextWithIds :: InteractiveContext -> [Id] -> InteractiveContext
 -- Just a specialised version
 extendInteractiveContextWithIds ictxt new_ids
   | null new_ids = ictxt
-  | otherwise    = ictxt { ic_mod_index  = ic_mod_index ictxt + 1
-                         , ic_tythings   = new_tythings ++ old_tythings
-                         , ic_rn_gbl_env = ic_rn_gbl_env ictxt `icExtendGblRdrEnv` new_tythings }
+  | otherwise
+  = ictxt { ic_mod_index  = ic_mod_index ictxt + 1
+          , ic_tythings   = new_tythings ++ ic_tythings ictxt
+          , ic_gre_cache  = ic_gre_cache ictxt `icExtendIcGblRdrEnv` new_tythings
+          }
   where
     new_tythings = map AnId new_ids
-    old_tythings = filterOut (shadowed_by new_ids) (ic_tythings ictxt)
-
-shadowed_by :: [Id] -> TyThing -> Bool
-shadowed_by ids = shadowed
-  where
-    -- Keep record selectors because they might be needed by HasField (#19322)
-    shadowed (AnId id) | isRecordSelector id = False
-    shadowed tything = getOccName tything `elemOccSet` new_occs
-    new_occs = mkOccSet (map getOccName ids)
 
 setInteractivePrintName :: InteractiveContext -> Name -> InteractiveContext
 setInteractivePrintName ic n = ic{ic_int_print = n}
 
-    -- ToDo: should not add Ids to the gbl env here
+icExtendIcGblRdrEnv :: IcGlobalRdrEnv -> [TyThing] -> IcGlobalRdrEnv
+icExtendIcGblRdrEnv igre tythings = IcGlobalRdrEnv
+    { igre_env        = icExtendGblRdrEnv False (igre_env igre)        tythings
+    , igre_prompt_env = icExtendGblRdrEnv True  (igre_prompt_env igre) tythings
+        -- Pass 'True' <=> drop names that are only available qualified.
+        -- This is done to maintain the invariant of Note [icReaderEnv recalculation]
+        -- that igre_prompt_env should only contain Names that are available unqualified.
+    }
 
--- | Add TyThings to the GlobalRdrEnv, earlier ones in the list shadowing
--- later ones, and shadowing existing entries in the GlobalRdrEnv.
-icExtendGblRdrEnv :: GlobalRdrEnv -> [TyThing] -> GlobalRdrEnv
-icExtendGblRdrEnv env tythings
+-- This is used by setContext in GHC.Runtime.Eval when the set of imports
+-- changes, and recalculates the GlobalRdrEnv. See Note [icReaderEnv recalculation]
+replaceImportEnv :: IcGlobalRdrEnv -> GlobalRdrEnv -> IcGlobalRdrEnv
+replaceImportEnv igre import_env = igre { igre_env = new_env }
+  where
+    import_env_shadowed = shadowNames False import_env (igre_prompt_env igre)
+    new_env = import_env_shadowed `plusGlobalRdrEnv` igre_prompt_env igre
+
+-- | Add 'TyThings' to the 'GlobalRdrEnv', earlier ones in the list shadowing
+-- later ones, and shadowing existing entries in the 'GlobalRdrEnv'.
+icExtendGblRdrEnv :: Bool -- ^ discard names that are only available qualified?
+                  -> GlobalRdrEnv -> [TyThing] -> GlobalRdrEnv
+icExtendGblRdrEnv drop_only_qualified env tythings
   = foldr add env tythings  -- Foldr makes things in the front of
                             -- the list shadow things at the back
   where
@@ -365,10 +430,10 @@ icExtendGblRdrEnv env tythings
        | is_sub_bndr thing
        = env
        | otherwise
-       = foldl' extendGlobalRdrEnv env1 (concatMap localGREsFromAvail avail)
+       = foldl' extendGlobalRdrEnv env1 new_gres
        where
-          env1  = shadowNames env (concatMap availGreNames avail)
-          avail = tyThingAvailInfo thing
+          new_gres = tyThingLocalGREs thing
+          env1     = shadowNames drop_only_qualified env $ mkGlobalRdrEnv new_gres
 
     -- Ugh! The new_tythings may include record selectors, since they
     -- are not implicit-ids, and must appear in the TypeEnv.  But they
@@ -381,7 +446,7 @@ icExtendGblRdrEnv env tythings
                              _            -> False
     is_sub_bndr _ = False
 
-substInteractiveContext :: InteractiveContext -> TCvSubst -> InteractiveContext
+substInteractiveContext :: InteractiveContext -> Subst -> InteractiveContext
 substInteractiveContext ictxt@InteractiveContext{ ic_tythings = tts } subst
   | isEmptyTCvSubst subst = ictxt
   | otherwise             = ictxt { ic_tythings = map subst_ty tts }
@@ -397,4 +462,3 @@ substInteractiveContext ictxt@InteractiveContext{ ic_tythings = tts } subst
 instance Outputable InteractiveImport where
   ppr (IIModule m) = char '*' <> ppr m
   ppr (IIDecl d)   = ppr d
-
