@@ -18,13 +18,14 @@ module Hadrian.Utilities (
     -- * File system operations
     copyFile, copyFileUntracked, createFileLink, fixFile,
     makeExecutable, moveFile, removeFile, createDirectory, copyDirectory,
-    moveDirectory, removeDirectory, removeFile_,
+    moveDirectory, removeDirectory, removeFile_, writeFileChangedBS,
+    findExecutable,
 
     -- * Diagnostic info
     Colour (..), ANSIColour (..), putColoured, shouldUseColor,
     BuildProgressColour, mkBuildProgressColour, putBuild,
     SuccessColour, mkSuccessColour, putSuccess,
-    FailureColour, mkFailureColour, putFailure,
+    FailureColour(..), red, mkFailureColour, putFailure,
     ProgressInfo (..), putProgressInfo,
     renderAction, renderActionNoOutput, renderProgram, renderLibrary, renderBox, renderUnicorn,
 
@@ -49,6 +50,7 @@ import Development.Shake.Classes
 import Development.Shake.FilePath
 import System.Environment (lookupEnv)
 
+import qualified Data.ByteString        as BS
 import qualified Control.Exception.Base as IO
 import qualified Data.HashMap.Strict    as Map
 import qualified System.Directory.Extra as IO
@@ -394,6 +396,25 @@ removeDirectory dir = do
     putProgressInfo $ "| Remove directory " ++ dir
     liftIO . whenM (IO.doesDirectoryExist dir) $ IO.removeDirectoryRecursive dir
 
+-- | Like Shake's 'writeFileChanged', but accepts a 'ByteString'.
+writeFileChangedBS :: FilePath -> BS.ByteString -> Action ()
+writeFileChangedBS name new = do
+    liftIO $ IO.createDirectoryIfMissing True $ takeDirectory name
+    exists <- liftIO $ IO.doesFileExist name
+    if exists
+      then do
+        old <- liftIO $ BS.readFile name
+        when (old /= new) $ do
+          liftIO $ removeFile_ name
+          do_write
+      else
+        do_write
+  where
+    do_write = do
+      putProgressInfo $ "| Write file " ++ name
+      liftIO $ BS.writeFile name new
+
+
 -- | Terminal output colours
 data Colour
     = Dull ANSIColour   -- ^ 8-bit ANSI colours
@@ -430,20 +451,20 @@ mkColour (Dull c) = colourCode c
 mkColour (Vivid c) = colourCode c ++ ";1"
 mkColour (Extended code) = "38;5;" ++ code
 
--- | A more colourful version of Shake's 'putNormal'.
+-- | A more colourful version of Shake's 'putInfo'.
 putColoured :: String -> String -> Action ()
 putColoured code msg = do
     useColour <- shakeColor <$> getShakeOptions
     if useColour
-        then putNormal $ "\ESC[" ++ code ++ "m" ++ msg ++ "\ESC[0m"
-        else putNormal msg
+        then putInfo $ "\ESC[" ++ code ++ "m" ++ msg ++ "\ESC[0m"
+        else putInfo msg
 
 newtype BuildProgressColour = BuildProgressColour String
     deriving Typeable
 
 -- | By default, Hadrian tries to figure out if the current terminal
 --   supports colors using this function. The default can be overridden
---   by suppling @--[no-]color@.
+--   by supplying @--[no-]color@.
 shouldUseColor :: IO Bool
 shouldUseColor =
   (&&) <$> IO.hIsTerminalDevice IO.stdout
@@ -554,22 +575,36 @@ renderCreateFileLink linkTarget link' = do
     where
         link = unifyPath link'
 
+-- | Render a multiline string, prefixing the first line with a header.
+renderMultiLineString :: String -> String -> [String]
+renderMultiLineString header string =
+    [ linePrefix index ++ line | (index, line) <- zip [0..] (lines string) ]
+  where
+    linePrefix :: Int -> String
+    linePrefix index
+        | index == 0 = header
+        | otherwise  = replicate (length header) ' '
+
+-- | Render a (possibly multiline) synopsis, making sure it ends with a dot.
+renderSynopsis :: String -> String -> [String]
+renderSynopsis header synopsis
+    | null synopsis = []
+    | otherwise = renderMultiLineString header (endWithADot synopsis)
+  where
+    endWithADot :: String -> String
+    endWithADot s = dropWhileEnd isPunctuation s ++ "."
+
 -- | Render the successful build of a program.
 renderProgram :: String -> String -> String -> String
 renderProgram name bin synopsis = renderBox $
-    [ "Successfully built program " ++ name
-    , "Executable: " ++ bin ] ++
-    [ "Program synopsis: " ++ endWithADot synopsis | not (null synopsis) ]
+    [ "Successfully built program " ++ name, "Executable: " ++ bin ] ++
+    renderSynopsis "Program synopsis: " synopsis
 
 -- | Render the successful build of a library.
 renderLibrary :: String -> String -> String -> String
 renderLibrary name lib synopsis = renderBox $
-    [ "Successfully built library " ++ name
-    , "Library: " ++ lib ] ++
-    [ "Library synopsis: " ++ endWithADot synopsis | not (null synopsis) ]
-
-endWithADot :: String -> String
-endWithADot s = dropWhileEnd isPunctuation s ++ "."
+    [ "Successfully built library " ++ name, "Library: " ++ lib ] ++
+    renderSynopsis "Library synopsis: " synopsis
 
 -- | Render the given set of lines in an ASCII box. The minimum width and
 -- whether to use Unicode symbols are hardcoded in the function's body.
@@ -634,3 +669,7 @@ renderUnicorn ls =
     ponyPadding = "                                            "
     boxLines :: [String]
     boxLines = ["", "", ""] ++ (lines . renderBox $ ls)
+
+-- Workaround for https://github.com/haskell/directory/issues/180
+findExecutable :: String -> IO (Maybe FilePath)
+findExecutable exe = IO.catch (IO.findExecutable exe) $ \(_ :: IO.IOException) -> pure Nothing
