@@ -6,6 +6,7 @@ module Rules.Generate (
     ) where
 
 import Development.Shake.FilePath
+import Data.Char (isSpace)
 import qualified Data.Set as Set
 import Base
 import qualified Context
@@ -38,7 +39,13 @@ primopsTxt :: Stage -> FilePath
 primopsTxt stage = buildDir (vanillaContext stage compiler) -/- "primops.txt"
 
 isGeneratedCmmFile :: FilePath -> Bool
-isGeneratedCmmFile file = takeBaseName file == "AutoApply"
+isGeneratedCmmFile file =
+  takeBaseName file `elem`
+    [ "AutoApply"
+    , "AutoApply_V16"
+    , "AutoApply_V32"
+    , "AutoApply_V64"
+    ]
 
 ghcPrimDependencies :: Expr [FilePath]
 ghcPrimDependencies = do
@@ -94,6 +101,7 @@ compilerDependencies = do
                   , "primop-vector-tys.hs-incl"
                   , "primop-vector-uniques.hs-incl"
                   , "primop-docs.hs-incl"
+                  , "primop-deprecations.hs-incl"
                   , "GHC/Platform/Constants.hs"
                   , "GHC/Settings/Config.hs"
                   ]
@@ -154,7 +162,22 @@ generatePackageCode context@(Context stage pkg _ _) = do
             path <- buildPath context
             let h = path -/- "include/DerivedConstants.h"
             need [h]
-            build $ target context GenApply [h] [file]
+            build $ target context (GenApply Nothing) [h] [file]
+        root -/- "**" -/- dir -/- "cmm/AutoApply_V16.cmm" %> \file -> do
+            path <- buildPath context
+            let h = path -/- "include/DerivedConstants.h"
+            need [h]
+            build $ target context (GenApply (Just 16)) [h] [file]
+        root -/- "**" -/- dir -/- "cmm/AutoApply_V32.cmm" %> \file -> do
+            path <- buildPath context
+            let h = path -/- "include/DerivedConstants.h"
+            need [h]
+            build $ target context (GenApply (Just 32)) [h] [file]
+        root -/- "**" -/- dir -/- "cmm/AutoApply_V64.cmm" %> \file -> do
+            path <- buildPath context
+            let h = path -/- "include/DerivedConstants.h"
+            need [h]
+            build $ target context (GenApply (Just 64)) [h] [file]
         root -/- "**" -/- dir -/- "include/ghcautoconf.h" %> \_ ->
             need . pure =<< pkgSetupConfigFile context
         root -/- "**" -/- dir -/- "include/ghcplatform.h" %> \_ ->
@@ -211,11 +234,15 @@ copyRules = do
         prefix -/- "ghc-interp.js"     <~ return "."
         prefix -/- "template-hsc.h" <~ return (pkgPath hsc2hs -/- "data")
 
+        prefix -/- "dyld.mjs" %> \file -> do
+            copyFile ("utils/jsffi" -/- makeRelative prefix file) file
+            makeExecutable file
+
         prefix -/- "post-link.mjs" %> \file -> do
             copyFile ("utils/jsffi" -/- makeRelative prefix file) file
             makeExecutable file
 
-        prefix -/- "prelude.js"        <~ pure "utils/jsffi"
+        prefix -/- "prelude.mjs"       <~ pure "utils/jsffi"
 
         prefix -/- "html/**"           <~ return "utils/haddock/haddock-api/resources"
         prefix -/- "latex/**"          <~ return "utils/haddock/haddock-api/resources"
@@ -234,8 +261,8 @@ generateRules = do
 
     forM_ allStages $ \stage -> do
         let prefix = root -/- stageString stage -/- "lib"
-            go gen file = generate file (semiEmptyTarget stage) gen
-        (prefix -/- "settings") %> go generateSettings
+            go gen file = generate file (semiEmptyTarget (succStage stage)) gen
+        (prefix -/- "settings") %> \out -> go (generateSettings out) out
 
   where
     file <~+ gen = file %> \out -> generate out emptyTarget gen >> makeExecutable out
@@ -278,11 +305,12 @@ runInterpolations (Interpolations mk_substs) input = do
 interpolateSetting :: String -> Setting -> Interpolations
 interpolateSetting name settng = interpolateVar name $ setting settng
 
--- | Interpolate the @ProjectVersion@ and @ProjectVersionMunged@ variables.
+-- | Interpolate the @ProjectVersion@, @ProjectVersionMunged@, and @ProjectVersionForLib@ variables.
 projectVersion :: Interpolations
 projectVersion = mconcat
     [ interpolateSetting "ProjectVersion" ProjectVersion
     , interpolateSetting "ProjectVersionMunged" ProjectVersionMunged
+    , interpolateSetting "ProjectVersionForLib" ProjectVersionForLib
     ]
 
 packageVersions :: Interpolations
@@ -306,14 +334,18 @@ escapedPkgName = map f . pkgName
     f '-'   = '_'
     f other = other
 
-templateRule :: FilePath -> Interpolations -> Rules ()
-templateRule outPath interps = do
+templateRuleFrom :: FilePath -> FilePath -> Interpolations -> Rules ()
+templateRuleFrom inPath outPath interps = do
     outPath %> \_ -> do
-        s <- readFile' (outPath <.> "in")
+        s <- readFile' inPath
         result <- runInterpolations interps s
         writeFile' outPath result
         putSuccess ("| Successfully generated " ++ outPath ++ " from its template")
 
+templateRule :: FilePath -> Interpolations -> Rules ()
+templateRule outPath =
+  templateRuleFrom (outPath <.> "in") outPath
+
 templateRules :: Rules ()
 templateRules = do
   templateRule "compiler/ghc.cabal" $ projectVersion
@@ -323,11 +355,26 @@ templateRules = do
   templateRule "utils/remote-iserv/remote-iserv.cabal" $ projectVersion
   templateRule "utils/runghc/runghc.cabal" $ projectVersion
   templateRule "libraries/ghc-boot/ghc-boot.cabal" $ projectVersion
-  templateRule "libraries/ghc-boot-th/ghc-boot-th.cabal" $ projectVersion
+  templateRule "libraries/ghc-boot-th/ghc-boot-th.cabal" $ mconcat
+    [ projectVersion
+    , interpolateVar "Suffix" $ pure ""
+    , interpolateVar "SourceRoot" $ pure "."
+    ]
+  templateRuleFrom "libraries/ghc-boot-th/ghc-boot-th.cabal.in"
+                   "libraries/ghc-boot-th-next/ghc-boot-th-next.cabal" $ mconcat
+    [ projectVersion
+    , interpolateVar "Suffix" $ pure "-next"
+    , interpolateVar "SourceRoot" $ pure "../ghc-boot-th"
+    ]
   templateRule "libraries/ghci/ghci.cabal" $ projectVersion
   templateRule "libraries/ghc-heap/ghc-heap.cabal" $ projectVersion
+  templateRule "libraries/ghc-internal/ghc-internal.cabal" $ projectVersion
+  templateRule "libraries/ghc-experimental/ghc-experimental.cabal" $ projectVersion
+  templateRule "libraries/base/base.cabal" $ projectVersion
   templateRule "utils/ghc-pkg/ghc-pkg.cabal" $ projectVersion
-  templateRule "libraries/template-haskell/template-haskell.cabal" $ projectVersion
+  templateRule "libraries/template-haskell/template-haskell.cabal" $ mconcat
+    [ projectVersion
+    ]
   templateRule "libraries/prologue.txt" $ packageVersions
   templateRule "rts/include/ghcversion.h" $ mconcat
     [ interpolateSetting "ProjectVersionInt" ProjectVersionInt
@@ -346,6 +393,57 @@ templateRules = do
     [ interpolateVar "LlvmMinVersion" $ replaceEq '.' ',' <$> setting LlvmMinVersion
     , interpolateVar "LlvmMaxVersion" $ replaceEq '.' ',' <$> setting LlvmMaxVersion
     ]
+  bindistRules
+
+bindistRules :: Rules ()
+bindistRules = do
+  templateRule ("mk" -/- "project.mk") $ mconcat
+    [ interpolateSetting "ProjectName" ProjectName
+    , interpolateSetting "ProjectVersion" ProjectVersion
+    , interpolateSetting "ProjectVersionInt" ProjectVersionInt
+    , interpolateSetting "ProjectPatchLevel" ProjectPatchLevel
+    , interpolateSetting "ProjectPatchLevel1" ProjectPatchLevel1
+    , interpolateSetting "ProjectPatchLevel2" ProjectPatchLevel2
+    , interpolateSetting "ProjectGitCommitId" ProjectGitCommitId
+
+    , interpolateVar "HostOS_CPP" $ fmap cppify $ interp $ queryHost queryOS
+
+    , interpolateVar "TargetPlatform" $ getTarget targetPlatformTriple
+    , interpolateVar "TargetPlatform_CPP" $ cppify <$> getTarget targetPlatformTriple
+    , interpolateVar "TargetArch_CPP" $ cppify <$> getTarget queryArch
+    , interpolateVar "TargetOS_CPP" $ cppify <$> getTarget queryOS
+    , interpolateVar "LLVMTarget" $ getTarget tgtLlvmTarget
+    ]
+  templateRule ("distrib" -/- "configure.ac") $ mconcat
+    [ interpolateSetting "ConfiguredEmsdkVersion" EmsdkVersion
+    , interpolateVar "CrossCompilePrefix" $ do
+        crossCompiling <- interp $ getFlag CrossCompiling
+        tpf <- setting TargetPlatformFull
+        pure $ if crossCompiling then tpf <> "-" else ""
+    , interpolateVar "LeadingUnderscore" $ yesNo <$> getTarget tgtSymbolsHaveLeadingUnderscore
+    , interpolateSetting "LlvmMaxVersion" LlvmMaxVersion
+    , interpolateSetting "LlvmMinVersion" LlvmMinVersion
+    , interpolateVar "LlvmTarget" $ getTarget tgtLlvmTarget
+    , interpolateSetting "ProjectVersion" ProjectVersion
+    , interpolateVar "SettingsUseDistroMINGW" $ settingsFileSetting ToolchainSetting_DistroMinGW
+    , interpolateVar "TablesNextToCode" $ yesNo <$> getTarget tgtTablesNextToCode
+    , interpolateVar "TargetHasLibm" $ lookupSystemConfig "target-has-libm"
+    , interpolateVar "TargetPlatform" $ getTarget targetPlatformTriple
+    , interpolateVar "TargetWordBigEndian" $ getTarget isBigEndian
+    , interpolateVar "TargetWordSize" $ getTarget wordSize
+    , interpolateVar "Unregisterised" $ yesNo <$> getTarget tgtUnregisterised
+    , interpolateVar "UseLibdw" $ fmap yesNo $ interp $ getFlag UseLibdw
+    , interpolateVar "UseLibffiForAdjustors" $ yesNo <$> getTarget tgtUseLibffiForAdjustors
+    , interpolateVar "GhcWithSMP" $ yesNo <$> targetSupportsSMP
+    ]
+  where
+    interp = interpretInContext (semiEmptyTarget Stage2)
+    getTarget = interp . queryTarget
+
+-- | Given a 'String' replace characters '.' and '-' by underscores ('_') so that
+-- the resulting 'String' is a valid C preprocessor identifier.
+cppify :: String -> String
+cppify = replaceEq '-' '_' . replaceEq '.' '_'
 
 
 -- Generators
@@ -355,21 +453,26 @@ templateRules = do
 ghcWrapper :: Stage -> Expr String
 ghcWrapper (Stage0 {}) = error "Stage0 GHC does not require a wrapper script to run."
 ghcWrapper stage  = do
-    dbPath  <- expr $ (</>) <$> topDirectory <*> packageDbPath (PackageDbLoc stage Final)
     ghcPath <- expr $ (</>) <$> topDirectory
                             <*> programPath (vanillaContext (predStage stage) ghc)
     return $ unwords $ map show $ [ ghcPath ]
-                               ++ (if stage == Stage1
-                                     then ["-no-global-package-db"
-                                          , "-package-env=-"
-                                          , "-package-db " ++ dbPath
-                                          ]
-                                     else [])
                                ++ [ "$@" ]
 
-generateSettings :: Expr String
-generateSettings = do
+generateSettings :: FilePath -> Expr String
+generateSettings settingsFile = do
     ctx <- getContext
+    stage <- getStage
+
+    package_db_path <- expr $ do
+      let get_pkg_db stg = packageDbPath (PackageDbLoc stg Final)
+      case stage of
+        Stage0 {} -> error "Unable to generate settings for stage0"
+        Stage1 -> get_pkg_db Stage1
+        Stage2 -> get_pkg_db Stage1
+        Stage3 -> get_pkg_db Stage2
+
+    let rel_pkg_db = makeRelativeNoSysLink (dropFileName settingsFile) package_db_path
+
     settings <- traverse sequence $
         [ ("C compiler command",   queryTarget ccPath)
         , ("C compiler flags",     queryTarget ccFlags)
@@ -381,6 +484,11 @@ generateSettings = do
         , ("CPP flags",           queryTarget cppFlags)
         , ("Haskell CPP command", queryTarget hsCppPath)
         , ("Haskell CPP flags",   queryTarget hsCppFlags)
+        , ("JavaScript CPP command", queryTarget jsCppPath)
+        , ("JavaScript CPP flags", queryTarget jsCppFlags)
+        , ("C-- CPP command", queryTarget cmmCppPath)
+        , ("C-- CPP flags",   queryTarget cmmCppFlags)
+        , ("C-- CPP supports -g0", queryTarget cmmCppSupportsG0')
         , ("ld supports compact unwind", queryTarget linkSupportsCompactUnwind)
         , ("ld supports filelist",       queryTarget linkSupportsFilelist)
         , ("ld supports single module",       queryTarget linkSupportsSingleModule)
@@ -414,13 +522,15 @@ generateSettings = do
         , ("LLVM llvm-as command", expr $ settingsFileSetting ToolchainSetting_LlvmAsCommand)
         , ("Use inplace MinGW toolchain", expr $ settingsFileSetting ToolchainSetting_DistroMinGW)
 
+        , ("target RTS linker only supports shared libraries", expr $ yesNo <$> targetRTSLinkerOnlySupportsSharedLibs)
         , ("Use interpreter", expr $ yesNo <$> ghcWithInterpreter)
         , ("Support SMP", expr $ yesNo <$> targetSupportsSMP)
-        , ("RTS ways", unwords . map show . Set.toList <$> getRtsWays)
+        , ("RTS ways", escapeArgs . map show . Set.toList <$> getRtsWays)
         , ("Tables next to code", queryTarget (yesNo . tgtTablesNextToCode))
         , ("Leading underscore",  queryTarget (yesNo . tgtSymbolsHaveLeadingUnderscore))
         , ("Use LibFFI", expr $ yesNo <$> useLibffiForAdjustors)
         , ("RTS expects libdw", yesNo <$> getFlag UseLibdw)
+        , ("Relative Global Package DB", pure rel_pkg_db)
         ]
     let showTuple (k, v) = "(" ++ show k ++ ", " ++ show v ++ ")"
     pure $ case settings of
@@ -431,30 +541,36 @@ generateSettings = do
             ++ ["]"]
   where
     ccPath  = prgPath . ccProgram . tgtCCompiler
-    ccFlags = unwords . prgFlags . ccProgram . tgtCCompiler
+    ccFlags = escapeArgs . prgFlags . ccProgram . tgtCCompiler
     cxxPath  = prgPath . cxxProgram . tgtCxxCompiler
-    cxxFlags = unwords . prgFlags . cxxProgram . tgtCxxCompiler
-    clinkFlags = unwords . prgFlags . ccLinkProgram . tgtCCompilerLink
+    cxxFlags = escapeArgs . prgFlags . cxxProgram . tgtCxxCompiler
+    clinkFlags = escapeArgs . prgFlags . ccLinkProgram . tgtCCompilerLink
     linkSupportsNoPie = yesNo . ccLinkSupportsNoPie . tgtCCompilerLink
     cppPath  = prgPath . cppProgram . tgtCPreprocessor
-    cppFlags = unwords . prgFlags . cppProgram . tgtCPreprocessor
+    cppFlags = escapeArgs . prgFlags . cppProgram . tgtCPreprocessor
     hsCppPath  = prgPath . hsCppProgram . tgtHsCPreprocessor
-    hsCppFlags = unwords . prgFlags . hsCppProgram . tgtHsCPreprocessor
+    hsCppFlags = escapeArgs . prgFlags . hsCppProgram . tgtHsCPreprocessor
+    jsCppPath  = maybe "" (prgPath . jsCppProgram) . tgtJsCPreprocessor
+    jsCppFlags = maybe "" (escapeArgs . prgFlags . jsCppProgram) . tgtJsCPreprocessor
+    cmmCppPath  = prgPath . cmmCppProgram . tgtCmmCPreprocessor
+    cmmCppFlags = escapeArgs . prgFlags . cmmCppProgram . tgtCmmCPreprocessor
+    cmmCppSupportsG0' = yesNo . cmmCppSupportsG0 . tgtCmmCPreprocessor
     mergeObjsPath  = maybe "" (prgPath . mergeObjsProgram) . tgtMergeObjs
-    mergeObjsFlags = maybe "" (unwords . prgFlags . mergeObjsProgram) . tgtMergeObjs
+    mergeObjsFlags = maybe "" (escapeArgs . prgFlags . mergeObjsProgram) . tgtMergeObjs
     linkSupportsSingleModule    = yesNo . ccLinkSupportsSingleModule . tgtCCompilerLink
     linkSupportsFilelist        = yesNo . ccLinkSupportsFilelist . tgtCCompilerLink
     linkSupportsCompactUnwind   = yesNo . ccLinkSupportsCompactUnwind . tgtCCompilerLink
     linkIsGnu                   = yesNo . ccLinkIsGnu . tgtCCompilerLink
     arPath  = prgPath . arMkArchive . tgtAr
-    arFlags = unwords . prgFlags . arMkArchive . tgtAr
+    arFlags = escapeArgs . prgFlags . arMkArchive . tgtAr
     arSupportsAtFile' = yesNo . arSupportsAtFile . tgtAr
     arSupportsDashL' = yesNo . arSupportsDashL . tgtAr
     ranlibPath  = maybe "" (prgPath . ranlibProgram) . tgtRanlib
-    isBigEndian = yesNo . (\case BigEndian -> True; LittleEndian -> False) . tgtEndianness
-    wordSize    = show . wordSize2Bytes . tgtWordSize
     mergeObjsSupportsResponseFiles' = maybe "NO" (yesNo . mergeObjsSupportsResponseFiles) . tgtMergeObjs
 
+isBigEndian, wordSize :: Toolchain.Target -> String
+isBigEndian = yesNo . (\case BigEndian -> True; LittleEndian -> False) . tgtEndianness
+wordSize    = show . wordSize2Bytes . tgtWordSize
 
 -- | Generate @Config.hs@ files.
 generateConfigHs :: Expr String
@@ -571,3 +687,19 @@ generatePlatformHostHs = do
         , "hostPlatformArchOS :: ArchOS"
         , "hostPlatformArchOS = ArchOS hostPlatformArch hostPlatformOS"
         ]
+
+-- | Just like 'GHC.ResponseFile.escapeArgs', but use spaces instead of newlines
+-- for splitting elements.
+escapeArgs :: [String] -> String
+escapeArgs = unwords . map escapeArg
+
+escapeArg :: String -> String
+escapeArg = reverse . foldl' escape []
+
+escape :: String -> Char -> String
+escape cs c
+  |    isSpace c
+    || '\\' == c
+    || '\'' == c
+    || '"'  == c = c:'\\':cs -- n.b., our caller must reverse the result
+  | otherwise    = c:cs
