@@ -29,8 +29,11 @@ splitSectionsArgs = do
     ) then
     ( mconcat
         [ builder (Ghc CompileHs) ? arg "-fsplit-sections"
+        , builder (Ghc CompileCWithGhc) ? arg "-fsplit-sections"
+        , builder (Ghc CompileCppWithGhc) ? arg "-fsplit-sections"
+        , builder (Cc CompileC) ? arg "-ffunction-sections" <> arg "-fdata-sections"
         , builder MergeObjects ? ifM (expr isWinTarget)
-            (pure ["-t", "driver/utils/merge_sections_pe.ld"])
-            (pure ["-t", "driver/utils/merge_sections.ld"])
+            (pure ["-T", "driver/utils/merge_sections_pe.ld"])
+            (pure ["-T", "driver/utils/merge_sections.ld"])
         ]
     ) else mempty
