@@ -1,12 +1,13 @@
+{-# LANGUAGE DataKinds #-}
 {-# LANGUAGE LambdaCase #-}
+{-# LANGUAGE RankNTypes #-}
+{-# LANGUAGE TypeApplications #-}
 
 module Distribution.Simple.GHC.Build.Link where
 
 import Distribution.Compat.Prelude
 import Prelude ()
 
-import Control.Exception (assert)
-import Control.Monad (forM_)
 import Control.Monad.IO.Class
 import qualified Data.ByteString.Lazy.Char8 as BS
 import qualified Data.Set as Set
@@ -35,15 +36,27 @@ import qualified Distribution.Simple.Program.Ar as Ar
 import Distribution.Simple.Program.GHC
 import qualified Distribution.Simple.Program.Ld as Ld
 import Distribution.Simple.Setup.Common
+import Distribution.Simple.Setup.Config
 import Distribution.Simple.Setup.Repl
 import Distribution.Simple.Utils
 import Distribution.System
 import Distribution.Types.ComponentLocalBuildInfo
 import Distribution.Utils.NubList
+import Distribution.Utils.Path
 import Distribution.Verbosity
 import Distribution.Version
+
 import System.Directory
+  ( createDirectoryIfMissing
+  , doesDirectoryExist
+  , doesFileExist
+  , removeFile
+  , renameFile
+  )
 import System.FilePath
+  ( isRelative
+  , replaceExtension
+  )
 
 -- | Links together the object files of the Haskell modules and extra sources
 -- using the context in which the component is being built.
@@ -54,129 +67,161 @@ linkOrLoadComponent
   -- ^ The configured GHC program that will be used for linking
   -> PackageDescription
   -- ^ The package description containing the component being built
-  -> [FilePath]
+  -> [SymbolicPath Pkg File]
   -- ^ The full list of extra build sources (all C, C++, Js,
   -- Asm, and Cmm sources), which were compiled to object
   -- files.
-  -> (FilePath, FilePath)
+  -> (SymbolicPath Pkg (Dir Artifacts), SymbolicPath Pkg (Dir Build))
   -- ^ The build target dir, and the target dir.
   -- See Note [Build Target Dir vs Target Dir] in Distribution.Simple.GHC.Build
-  -> (Set.Set BuildWay, BuildWay -> GhcOptions)
+  -> ((Bool -> [BuildWay], Bool -> BuildWay, BuildWay), BuildWay -> GhcOptions)
   -- ^ The set of build ways wanted based on the user opts, and a function to
   -- convert a build way into the set of ghc options that were used to build
   -- that way.
   -> PreBuildComponentInputs
   -- ^ The context and component being built in it.
   -> IO ()
-linkOrLoadComponent ghcProg pkg_descr extraSources (buildTargetDir, targetDir) (wantedWays, buildOpts) pbci = do
-  let
-    verbosity = buildVerbosity pbci
-    target = targetInfo pbci
-    component = buildComponent pbci
-    what = buildingWhat pbci
-    lbi = localBuildInfo pbci
-    bi = buildBI pbci
-    clbi = buildCLBI pbci
+linkOrLoadComponent
+  ghcProg
+  pkg_descr
+  extraSources
+  (buildTargetDir, targetDir)
+  ((wantedLibWays, wantedFLibWay, wantedExeWay), buildOpts)
+  pbci = do
+    let
+      verbosity = buildVerbosity pbci
+      target = targetInfo pbci
+      component = buildComponent pbci
+      what = buildingWhat pbci
+      lbi = localBuildInfo pbci
+      bi = buildBI pbci
+      clbi = buildCLBI pbci
+      isIndef = componentIsIndefinite clbi
+      mbWorkDir = mbWorkDirLBI lbi
 
-  -- ensure extra lib dirs exist before passing to ghc
-  cleanedExtraLibDirs <- liftIO $ filterM doesDirectoryExist (extraLibDirs bi)
-  cleanedExtraLibDirsStatic <- liftIO $ filterM doesDirectoryExist (extraLibDirsStatic bi)
+      -- See Note [Symbolic paths] in Distribution.Utils.Path
+      i = interpretSymbolicPathLBI lbi
 
-  let
-    extraSourcesObjs = map (`replaceExtension` objExtension) extraSources
+    -- ensure extra lib dirs exist before passing to ghc
+    cleanedExtraLibDirs <- liftIO $ filterM (doesDirectoryExist . i) (extraLibDirs bi)
+    cleanedExtraLibDirsStatic <- liftIO $ filterM (doesDirectoryExist . i) (extraLibDirsStatic bi)
 
-    -- TODO: Shouldn't we use withStaticLib for libraries and something else
-    -- for foreign libs in the three cases where we use `withFullyStaticExe` below?
-    linkerOpts rpaths =
-      mempty
-        { ghcOptLinkOptions =
-            PD.ldOptions bi
-              ++ [ "-static"
-                 | withFullyStaticExe lbi
-                 ]
-              -- Pass extra `ld-options` given
-              -- through to GHC's linker.
-              ++ maybe
-                []
-                programOverrideArgs
-                (lookupProgram ldProgram (withPrograms lbi))
-        , ghcOptLinkLibs =
-            if withFullyStaticExe lbi
-              then extraLibsStatic bi
-              else extraLibs bi
-        , ghcOptLinkLibPath =
-            toNubListR $
+    let
+      extraSourcesObjs :: [RelativePath Artifacts File]
+      extraSourcesObjs =
+        [ makeRelativePathEx $ getSymbolicPath src `replaceExtension` objExtension
+        | src <- extraSources
+        ]
+
+      -- TODO: Shouldn't we use withStaticLib for libraries and something else
+      -- for foreign libs in the three cases where we use `withFullyStaticExe` below?
+      linkerOpts rpaths =
+        mempty
+          { ghcOptLinkOptions =
+              PD.ldOptions bi
+                ++ [ "-static"
+                   | withFullyStaticExe lbi
+                   ]
+                -- Pass extra `ld-options` given
+                -- through to GHC's linker.
+                ++ maybe
+                  []
+                  programOverrideArgs
+                  (lookupProgram ldProgram (withPrograms lbi))
+          , ghcOptLinkLibs =
               if withFullyStaticExe lbi
-                then cleanedExtraLibDirsStatic
-                else cleanedExtraLibDirs
-        , ghcOptLinkFrameworks = toNubListR $ PD.frameworks bi
-        , ghcOptLinkFrameworkDirs = toNubListR $ PD.extraFrameworkDirs bi
-        , ghcOptInputFiles = toNubListR [buildTargetDir </> x | x <- extraSourcesObjs]
-        , ghcOptNoLink = Flag False
-        , ghcOptRPaths = rpaths
-        }
-  case what of
-    BuildRepl replFlags -> liftIO $ do
-      let
-        -- For repl we use the vanilla (static) ghc options
-        staticOpts = buildOpts StaticWay
-        replOpts =
-          staticOpts
-            { -- Repl options use Static as the base, but doesn't need to pass -static.
-              -- However, it maybe should, for uniformity.
-              ghcOptDynLinkMode = NoFlag
-            , ghcOptExtra =
-                Internal.filterGhciFlags
-                  (ghcOptExtra staticOpts)
-                  <> replOptionsFlags (replReplOptions replFlags)
-            , ghcOptInputModules = replNoLoad (replReplOptions replFlags) (ghcOptInputModules staticOpts)
-            , ghcOptInputFiles = replNoLoad (replReplOptions replFlags) (ghcOptInputFiles staticOpts)
-            }
-            -- For a normal compile we do separate invocations of ghc for
-            -- compiling as for linking. But for repl we have to do just
-            -- the one invocation, so that one has to include all the
-            -- linker stuff too, like -l flags and any .o files from C
-            -- files etc.
-            --
-            -- TODO: The repl doesn't use the runtime paths from linkerOpts
-            -- (ghcOptRPaths), which looks like a bug. After the refactor we
-            -- can fix this.
-            `mappend` linkerOpts mempty
-            `mappend` mempty
-              { ghcOptMode = toFlag GhcModeInteractive
-              , ghcOptOptimisation = toFlag GhcNoOptimisation
+                then extraLibsStatic bi
+                else extraLibs bi
+          , ghcOptLinkLibPath =
+              toNubListR $
+                if withFullyStaticExe lbi
+                  then cleanedExtraLibDirsStatic
+                  else cleanedExtraLibDirs
+          , ghcOptLinkFrameworks = toNubListR $ map getSymbolicPath $ PD.frameworks bi
+          , ghcOptLinkFrameworkDirs = toNubListR $ PD.extraFrameworkDirs bi
+          , ghcOptInputFiles =
+              toNubListR
+                [ coerceSymbolicPath $ buildTargetDir </> obj
+                | obj <- extraSourcesObjs
+                ]
+          , ghcOptNoLink = Flag False
+          , ghcOptRPaths = rpaths
+          }
+
+    case what of
+      BuildRepl replFlags -> liftIO $ do
+        let
+          -- For repl we use the vanilla (static) ghc options
+          staticOpts = buildOpts StaticWay
+          replOpts =
+            staticOpts
+              { -- Repl options use Static as the base, but doesn't need to pass -static.
+                -- However, it maybe should, for uniformity.
+                ghcOptDynLinkMode = NoFlag
+              , ghcOptExtra =
+                  Internal.filterGhciFlags
+                    (ghcOptExtra staticOpts)
+                    <> replOptionsFlags (replReplOptions replFlags)
+              }
+              -- For a normal compile we do separate invocations of ghc for
+              -- compiling as for linking. But for repl we have to do just
+              -- the one invocation, so that one has to include all the
+              -- linker stuff too, like -l flags and any .o files from C
+              -- files etc.
+              --
+              -- TODO: The repl doesn't use the runtime paths from linkerOpts
+              -- (ghcOptRPaths), which looks like a bug. After the refactor we
+              -- can fix this.
+              `mappend` linkerOpts mempty
+              `mappend` mempty
+                { ghcOptMode = toFlag GhcModeInteractive
+                , ghcOptOptimisation = toFlag GhcNoOptimisation
+                }
+          replOpts_final =
+            replOpts
+              { ghcOptInputModules = replNoLoad (replReplOptions replFlags) (ghcOptInputModules replOpts)
+              , ghcOptInputFiles = replNoLoad (replReplOptions replFlags) (ghcOptInputFiles replOpts)
               }
 
-      -- TODO: problem here is we need the .c files built first, so we can load them
-      -- with ghci, but .c files can depend on .h files generated by ghc by ffi
-      -- exports.
-      when (case component of CLib lib -> null (allLibModules lib clbi); _ -> False) $
-        warn verbosity "No exposed modules"
-      runReplOrWriteFlags ghcProg lbi replFlags replOpts (pkgName (PD.package pkg_descr)) target
-    _otherwise ->
-      let
-        runGhcProg = runGHC verbosity ghcProg comp platform
-        platform = hostPlatform lbi
-        comp = compiler lbi
-       in
-        when (not $ componentIsIndefinite clbi) $ do
-          -- If not building dynamically, we don't pass any runtime paths.
-          rpaths <- if DynWay `Set.member` wantedWays then getRPaths pbci else return (toNubListR [])
-          liftIO $ do
-            info verbosity "Linking..."
-            let linkExeLike name = linkExecutable (linkerOpts rpaths) (wantedWays, buildOpts) targetDir name runGhcProg lbi
-            case component of
-              CLib lib -> linkLibrary buildTargetDir cleanedExtraLibDirs pkg_descr verbosity runGhcProg lib lbi clbi extraSources rpaths wantedWays
-              CFLib flib -> linkFLib flib bi lbi (linkerOpts rpaths) (wantedWays, buildOpts) targetDir runGhcProg
-              CExe exe -> linkExeLike (exeName exe)
-              CTest test -> linkExeLike (testName test)
-              CBench bench -> linkExeLike (benchmarkName bench)
+        -- TODO: problem here is we need the .c files built first, so we can load them
+        -- with ghci, but .c files can depend on .h files generated by ghc by ffi
+        -- exports.
+        when (case component of CLib lib -> null (allLibModules lib clbi); _ -> False) $
+          warn verbosity "No exposed modules"
+        runReplOrWriteFlags ghcProg lbi replFlags replOpts_final (pkgName (PD.package pkg_descr)) target
+      _otherwise ->
+        let
+          runGhcProg = runGHC verbosity ghcProg comp platform mbWorkDir
+          platform = hostPlatform lbi
+          comp = compiler lbi
+          get_rpaths ways =
+            if DynWay `Set.member` ways then getRPaths pbci else return (toNubListR [])
+         in
+          when (not $ componentIsIndefinite clbi) $ do
+            -- If not building dynamically, we don't pass any runtime paths.
+            liftIO $ do
+              info verbosity "Linking..."
+              let linkExeLike name = do
+                    rpaths <- get_rpaths (Set.singleton wantedExeWay)
+                    linkExecutable (linkerOpts rpaths) (wantedExeWay, buildOpts) targetDir name runGhcProg lbi
+              case component of
+                CLib lib -> do
+                  let libWays = wantedLibWays isIndef
+                  rpaths <- get_rpaths (Set.fromList libWays)
+                  linkLibrary buildTargetDir cleanedExtraLibDirs pkg_descr verbosity runGhcProg lib lbi clbi extraSources rpaths libWays
+                CFLib flib -> do
+                  let flib_way = wantedFLibWay (withDynFLib flib)
+                  rpaths <- get_rpaths (Set.singleton flib_way)
+                  linkFLib flib bi lbi (linkerOpts rpaths) (flib_way, buildOpts) targetDir runGhcProg
+                CExe exe -> linkExeLike (exeName exe)
+                CTest test -> linkExeLike (testName test)
+                CBench bench -> linkExeLike (benchmarkName bench)
 
 -- | Link a library component
 linkLibrary
-  :: FilePath
+  :: SymbolicPath Pkg (Dir Artifacts)
   -- ^ The library target build directory
-  -> [FilePath]
+  -> [SymbolicPath Pkg (Dir Lib)]
   -- ^ The list of extra lib dirs that exist (aka "cleaned")
   -> PackageDescription
   -- ^ The package description containing this library
@@ -186,15 +231,18 @@ linkLibrary
   -> Library
   -> LocalBuildInfo
   -> ComponentLocalBuildInfo
-  -> [FilePath]
+  -> [SymbolicPath Pkg File]
   -- ^ Extra build sources (that were compiled to objects)
   -> NubListR FilePath
   -- ^ A list with the runtime-paths (rpaths), or empty if not linking dynamically
-  -> Set.Set BuildWay
+  -> [BuildWay]
   -- ^ Wanted build ways and corresponding build options
   -> IO ()
 linkLibrary buildTargetDir cleanedExtraLibDirs pkg_descr verbosity runGhcProg lib lbi clbi extraSources rpaths wantedWays = do
   let
+    common = configCommonFlags $ configFlags lbi
+    mbWorkDir = flagToMaybe $ setupWorkingDir common
+
     compiler_id = compilerId comp
     comp = compiler lbi
     ghcVersion = compilerVersion comp
@@ -202,16 +250,19 @@ linkLibrary buildTargetDir cleanedExtraLibDirs pkg_descr verbosity runGhcProg li
     uid = componentUnitId clbi
     libBi = libBuildInfo lib
     Platform _hostArch hostOS = hostPlatform lbi
-    vanillaLibFilePath = buildTargetDir </> mkLibName uid
-    profileLibFilePath = buildTargetDir </> mkProfLibName uid
+    vanillaLibFilePath = buildTargetDir </> makeRelativePathEx (mkLibName uid)
+    profileLibFilePath = buildTargetDir </> makeRelativePathEx (mkProfLibName uid)
     sharedLibFilePath =
       buildTargetDir
-        </> mkSharedLibName (hostPlatform lbi) compiler_id uid
+        </> makeRelativePathEx (mkSharedLibName (hostPlatform lbi) compiler_id uid)
+    profSharedLibFilePath =
+      buildTargetDir
+        </> makeRelativePathEx (mkProfSharedLibName (hostPlatform lbi) compiler_id uid)
     staticLibFilePath =
       buildTargetDir
-        </> mkStaticLibName (hostPlatform lbi) compiler_id uid
-    ghciLibFilePath = buildTargetDir </> Internal.mkGHCiLibName uid
-    ghciProfLibFilePath = buildTargetDir </> Internal.mkGHCiProfLibName uid
+        </> makeRelativePathEx (mkStaticLibName (hostPlatform lbi) compiler_id uid)
+    ghciLibFilePath = buildTargetDir </> makeRelativePathEx (Internal.mkGHCiLibName uid)
+    ghciProfLibFilePath = buildTargetDir </> makeRelativePathEx (Internal.mkGHCiProfLibName uid)
     libInstallPath =
       libdir $
         absoluteComponentInstallDirs
@@ -222,7 +273,11 @@ linkLibrary buildTargetDir cleanedExtraLibDirs pkg_descr verbosity runGhcProg li
     sharedLibInstallPath =
       libInstallPath
         </> mkSharedLibName (hostPlatform lbi) compiler_id uid
+    profSharedLibInstallPath =
+      libInstallPath
+        </> mkProfSharedLibName (hostPlatform lbi) compiler_id uid
 
+    getObjFiles :: BuildWay -> IO [SymbolicPath Pkg File]
     getObjFiles way =
       mconcat
         [ Internal.getHaskellObjects
@@ -233,20 +288,34 @@ linkLibrary buildTargetDir cleanedExtraLibDirs pkg_descr verbosity runGhcProg li
             buildTargetDir
             (buildWayPrefix way ++ objExtension)
             True
-        , pure $
-            map (buildTargetDir </>) $
-              map ((`replaceExtension` (buildWayPrefix way ++ objExtension))) extraSources
+        , pure $ map (srcObjPath way) extraSources
         , catMaybes
             <$> sequenceA
-              [ findFileWithExtension
+              [ findFileCwdWithExtension
+                mbWorkDir
                 [Suffix $ buildWayPrefix way ++ objExtension]
                 [buildTargetDir]
-                (ModuleName.toFilePath x ++ "_stub")
+                xPath
               | ghcVersion < mkVersion [7, 2] -- ghc-7.2+ does not make _stub.o files
               , x <- allLibModules lib clbi
+              , let xPath :: RelativePath Artifacts File
+                    xPath = makeRelativePathEx $ ModuleName.toFilePath x ++ "_stub"
               ]
         ]
 
+    -- Get the @.o@ path from a source path (e.g. @.hs@),
+    -- in the library target build directory.
+    srcObjPath :: BuildWay -> SymbolicPath Pkg File -> SymbolicPath Pkg File
+    srcObjPath way srcPath =
+      case symbolicPathRelative_maybe objPath of
+        -- Absolute path: should already be in the target build directory
+        -- (e.g. a preprocessed file)
+        -- TODO: assert this?
+        Nothing -> objPath
+        Just objRelPath -> coerceSymbolicPath buildTargetDir </> objRelPath
+      where
+        objPath = srcPath `replaceExtensionSymbolicPath` (buildWayPrefix way ++ objExtension)
+
     -- I'm fairly certain that, just like the executable, we can keep just the
     -- module input list, and point to the right sources dir (as is already
     -- done), and GHC will pick up the right suffix (p_ for profile, dyn_ when
@@ -291,11 +360,12 @@ linkLibrary buildTargetDir cleanedExtraLibDirs pkg_descr verbosity runGhcProg li
     -- After the relocation lib is created we invoke ghc -shared
     -- with the dependencies spelled out as -package arguments
     -- and ghc invokes the linker with the proper library paths
+    ghcSharedLinkArgs :: [SymbolicPath Pkg File] -> GhcOptions
     ghcSharedLinkArgs dynObjectFiles =
       ghcBaseLinkArgs
         { ghcOptShared = toFlag True
         , ghcOptDynLinkMode = toFlag GhcDynamicOnly
-        , ghcOptInputFiles = toNubListR dynObjectFiles
+        , ghcOptInputFiles = toNubListR $ map coerceSymbolicPath dynObjectFiles
         , ghcOptOutputFile = toFlag sharedLibFilePath
         , -- For dynamic libs, Mac OS/X needs to know the install location
           -- at build time. This only applies to GHC < 7.8 - see the
@@ -307,7 +377,33 @@ linkLibrary buildTargetDir cleanedExtraLibDirs pkg_descr verbosity runGhcProg li
               else mempty
         , ghcOptLinkLibs = extraLibs libBi
         , ghcOptLinkLibPath = toNubListR $ cleanedExtraLibDirs
-        , ghcOptLinkFrameworks = toNubListR $ PD.frameworks libBi
+        , ghcOptLinkFrameworks = toNubListR $ map getSymbolicPath $ PD.frameworks libBi
+        , ghcOptLinkFrameworkDirs =
+            toNubListR $ PD.extraFrameworkDirs libBi
+        , ghcOptRPaths = rpaths
+        }
+    ghcProfSharedLinkArgs pdynObjectFiles =
+      ghcBaseLinkArgs
+        { ghcOptShared = toFlag True
+        , ghcOptProfilingMode = toFlag True
+        , ghcOptProfilingAuto =
+            Internal.profDetailLevelFlag
+              True
+              (withProfLibDetail lbi)
+        , ghcOptDynLinkMode = toFlag GhcDynamicOnly
+        , ghcOptInputFiles = toNubListR pdynObjectFiles
+        , ghcOptOutputFile = toFlag profSharedLibFilePath
+        , -- For dynamic libs, Mac OS/X needs to know the install location
+          -- at build time. This only applies to GHC < 7.8 - see the
+          -- discussion in #1660.
+          ghcOptDylibName =
+            if hostOS == OSX
+              && ghcVersion < mkVersion [7, 8]
+              then toFlag profSharedLibInstallPath
+              else mempty
+        , ghcOptLinkLibs = extraLibs libBi
+        , ghcOptLinkLibPath = toNubListR $ cleanedExtraLibDirs
+        , ghcOptLinkFrameworks = toNubListR $ map getSymbolicPath $ PD.frameworks libBi
         , ghcOptLinkFrameworkDirs =
             toNubListR $ PD.extraFrameworkDirs libBi
         , ghcOptRPaths = rpaths
@@ -315,7 +411,7 @@ linkLibrary buildTargetDir cleanedExtraLibDirs pkg_descr verbosity runGhcProg li
     ghcStaticLinkArgs staticObjectFiles =
       ghcBaseLinkArgs
         { ghcOptStaticLib = toFlag True
-        , ghcOptInputFiles = toNubListR staticObjectFiles
+        , ghcOptInputFiles = toNubListR $ map coerceSymbolicPath staticObjectFiles
         , ghcOptOutputFile = toFlag staticLibFilePath
         , ghcOptLinkLibs = extraLibs libBi
         , -- TODO: Shouldn't this use cleanedExtraLibDirsStatic instead?
@@ -325,6 +421,7 @@ linkLibrary buildTargetDir cleanedExtraLibDirs pkg_descr verbosity runGhcProg li
   staticObjectFiles <- getObjFiles StaticWay
   profObjectFiles <- getObjFiles ProfWay
   dynamicObjectFiles <- getObjFiles DynWay
+  profDynamicObjectFiles <- getObjFiles ProfDynWay
 
   let
     linkWay = \case
@@ -338,6 +435,8 @@ linkLibrary buildTargetDir cleanedExtraLibDirs pkg_descr verbosity runGhcProg li
             ldProg
             ghciProfLibFilePath
             profObjectFiles
+      ProfDynWay -> do
+        runGhcProg $ ghcProfSharedLinkArgs profDynamicObjectFiles
       DynWay -> do
         runGhcProg $ ghcSharedLinkArgs dynamicObjectFiles
       StaticWay -> do
@@ -367,10 +466,10 @@ linkLibrary buildTargetDir cleanedExtraLibDirs pkg_descr verbosity runGhcProg li
 linkExecutable
   :: (GhcOptions)
   -- ^ The linker-specific GHC options
-  -> (Set.Set BuildWay, BuildWay -> GhcOptions)
+  -> (BuildWay, BuildWay -> GhcOptions)
   -- ^ The wanted build ways and corresponding GhcOptions that were
   -- used to compile the modules in that way.
-  -> FilePath
+  -> SymbolicPath Pkg (Dir Build)
   -- ^ The target dir (2024-01:note: not the same as build target
   -- dir, see Note [Build Target Dir vs Target Dir] in Distribution.Simple.GHC.Build)
   -> UnqualComponentName
@@ -379,28 +478,27 @@ linkExecutable
   -- ^ Run the configured GHC program
   -> LocalBuildInfo
   -> IO ()
-linkExecutable linkerOpts (wantedWays, buildOpts) targetDir targetName runGhcProg lbi = do
-  -- When building an executable, we should only "want" one build way.
-  assert (Set.size wantedWays == 1) $
-    forM_ wantedWays $ \way -> do
-      let baseOpts = buildOpts way
-          linkOpts =
-            baseOpts
-              `mappend` linkerOpts
-              `mappend` mempty
-                { -- If there are no input Haskell files we pass -no-hs-main, and
-                  -- assume there is a main function in another non-haskell object
-                  ghcOptLinkNoHsMain = toFlag (ghcOptInputFiles baseOpts == mempty && ghcOptInputScripts baseOpts == mempty)
-                }
-          comp = compiler lbi
+linkExecutable linkerOpts (way, buildOpts) targetDir targetName runGhcProg lbi = do
+  let baseOpts = buildOpts way
+      linkOpts =
+        baseOpts
+          `mappend` linkerOpts
+          `mappend` mempty
+            { -- If there are no input Haskell files we pass -no-hs-main, and
+              -- assume there is a main function in another non-haskell object
+              ghcOptLinkNoHsMain = toFlag (ghcOptInputFiles baseOpts == mempty && ghcOptInputScripts baseOpts == mempty)
+            }
+      comp = compiler lbi
 
-      -- Work around old GHCs not relinking in this
-      -- situation, see #3294
-      let target = targetDir </> exeTargetName (hostPlatform lbi) targetName
-      when (compilerVersion comp < mkVersion [7, 7]) $ do
-        e <- doesFileExist target
-        when e (removeFile target)
-      runGhcProg linkOpts{ghcOptOutputFile = toFlag target}
+  -- Work around old GHCs not relinking in this
+  -- situation, see #3294
+  let target =
+        targetDir </> makeRelativePathEx (exeTargetName (hostPlatform lbi) targetName)
+  when (compilerVersion comp < mkVersion [7, 7]) $ do
+    let targetPath = interpretSymbolicPathLBI lbi target
+    e <- doesFileExist targetPath
+    when e (removeFile targetPath)
+  runGhcProg linkOpts{ghcOptOutputFile = toFlag target}
 
 -- | Link a foreign library component
 linkFLib
@@ -409,16 +507,16 @@ linkFLib
   -> LocalBuildInfo
   -> (GhcOptions)
   -- ^ The linker-specific GHC options
-  -> (Set.Set BuildWay, BuildWay -> GhcOptions)
+  -> (BuildWay, BuildWay -> GhcOptions)
   -- ^ The wanted build ways and corresponding GhcOptions that were
   -- used to compile the modules in that way.
-  -> FilePath
+  -> SymbolicPath Pkg (Dir Build)
   -- ^ The target dir (2024-01:note: not the same as build target
   -- dir, see Note [Build Target Dir vs Target Dir] in Distribution.Simple.GHC.Build)
   -> (GhcOptions -> IO ())
   -- ^ Run the configured GHC program
   -> IO ()
-linkFLib flib bi lbi linkerOpts (wantedWays, buildOpts) targetDir runGhcProg = do
+linkFLib flib bi lbi linkerOpts (way, buildOpts) targetDir runGhcProg = do
   let
     comp = compiler lbi
 
@@ -432,7 +530,7 @@ linkFLib flib bi lbi linkerOpts (wantedWays, buildOpts) targetDir runGhcProg = d
       | otherwise =
           mempty
             { ghcOptLinkLibs = rtsOptLinkLibs
-            , ghcOptLinkLibPath = toNubListR $ rtsLibPaths rtsInfo
+            , ghcOptLinkLibPath = toNubListR $ map makeSymbolicPath $ rtsLibPaths rtsInfo
             }
       where
         threaded = hasThreaded bi
@@ -450,8 +548,8 @@ linkFLib flib bi lbi linkerOpts (wantedWays, buildOpts) targetDir runGhcProg = d
                   else statRtsVanillaLib (rtsStaticInfo rtsInfo)
           ]
 
-    linkOpts :: BuildWay -> GhcOptions
-    linkOpts way = case foreignLibType flib of
+    linkOpts :: GhcOptions
+    linkOpts = case foreignLibType flib of
       ForeignLibNativeShared ->
         (buildOpts way)
           `mappend` linkerOpts
@@ -460,7 +558,7 @@ linkFLib flib bi lbi linkerOpts (wantedWays, buildOpts) targetDir runGhcProg = d
             { ghcOptLinkNoHsMain = toFlag True
             , ghcOptShared = toFlag True
             , ghcOptFPic = toFlag True
-            , ghcOptLinkModDefFiles = toNubListR $ foreignLibModDefFile flib
+            , ghcOptLinkModDefFiles = toNubListR $ fmap getSymbolicPath $ foreignLibModDefFile flib
             }
       ForeignLibNativeStatic ->
         -- this should be caught by buildFLib
@@ -473,11 +571,10 @@ linkFLib flib bi lbi linkerOpts (wantedWays, buildOpts) targetDir runGhcProg = d
   -- soname on supported platforms.  See also the note for
   -- @flibBuildName@.
   let buildName = flibBuildName lbi flib
-  -- There should not be more than one wanted way when building an flib
-  assert (Set.size wantedWays == 1) $
-    forM_ wantedWays $ \way -> do
-      runGhcProg (linkOpts way){ghcOptOutputFile = toFlag (targetDir </> buildName)}
-      renameFile (targetDir </> buildName) (targetDir </> flibTargetName lbi flib)
+  let outFile = targetDir </> makeRelativePathEx buildName
+  runGhcProg linkOpts{ghcOptOutputFile = toFlag outFile}
+  let i = interpretSymbolicPathLBI lbi
+  renameFile (i outFile) (i targetDir </> flibTargetName lbi flib)
 
 -- | Calculate the RPATHs for the component we are building.
 --
@@ -534,7 +631,9 @@ getRPaths pbci = do
             OSX -> "@loader_path"
             _ -> "$ORIGIN"
           relPath p = if isRelative p then hostPref </> p else p
-          rpaths = toNubListR (map relPath libraryPaths) <> toNubListR (extraLibDirs bi)
+          rpaths =
+            toNubListR (map relPath libraryPaths)
+              <> toNubListR (map getSymbolicPath $ extraLibDirs bi)
       return rpaths
     else return mempty
 
@@ -628,18 +727,24 @@ runReplOrWriteFlags ghcProg lbi rflags ghcOpts pkg_name target =
       clbi = targetCLBI target
       comp = compiler lbi
       platform = hostPlatform lbi
+      common = configCommonFlags $ configFlags lbi
+      mbWorkDir = mbWorkDirLBI lbi
+      verbosity = fromFlag $ setupVerbosity common
    in case replOptionsFlagOutput (replReplOptions rflags) of
-        NoFlag -> runGHC (fromFlag $ replVerbosity rflags) ghcProg comp platform ghcOpts
+        NoFlag -> runGHC verbosity ghcProg comp platform mbWorkDir ghcOpts
         Flag out_dir -> do
-          src_dir <- getCurrentDirectory
           let uid = componentUnitId clbi
               this_unit = prettyShow uid
-              reexported_modules = [mn | LibComponentLocalBuildInfo{} <- [clbi], IPI.ExposedModule mn (Just{}) <- componentExposedModules clbi]
+              reexported_modules =
+                [ mn | LibComponentLocalBuildInfo{componentExposedModules = exposed_mods} <- [clbi], IPI.ExposedModule mn (Just{}) <- exposed_mods
+                ]
               hidden_modules = otherModules bi
               extra_opts =
                 concat $
                   [ ["-this-package-name", prettyShow pkg_name]
-                  , ["-working-dir", src_dir]
+                  , case mbWorkDir of
+                      Nothing -> []
+                      Just wd -> ["-working-dir", getSymbolicPath wd]
                   ]
                     ++ [ ["-reexported-module", prettyShow m] | m <- reexported_modules
                        ]
@@ -655,7 +760,9 @@ runReplOrWriteFlags ghcProg lbi rflags ghcOpts pkg_name target =
           writeFileAtomic (out_dir </> this_unit) $
             BS.pack $
               escapeArgs $
-                extra_opts ++ renderGhcOptions comp platform (ghcOpts{ghcOptMode = NoFlag})
+                extra_opts
+                  ++ renderGhcOptions comp platform (ghcOpts{ghcOptMode = NoFlag})
+                  ++ programOverrideArgs ghcProg
 
 replNoLoad :: Ord a => ReplOptions -> NubListR a -> NubListR a
 replNoLoad replFlags l
