@@ -1,4 +1,5 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE DataKinds #-}
 {-# LANGUAGE LambdaCase #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 
@@ -24,6 +25,7 @@ import Distribution.Client.Setup
   ( ActAsSetupFlags (..)
   , BuildFlags (..)
   , CheckFlags (..)
+  , CommonSetupFlags (..)
   , ConfigExFlags (..)
   , ConfigFlags (..)
   , FetchFlags (..)
@@ -49,6 +51,7 @@ import Distribution.Client.Setup
   , defaultConfigExFlags
   , defaultInstallFlags
   , fetchCommand
+  , filterCommonFlags
   , formatCommand
   , freezeCommand
   , genBoundsCommand
@@ -83,7 +86,6 @@ import Distribution.Simple.Setup
   , RegisterFlags (..)
   , ReplFlags (..)
   , TestFlags (..)
-  , configAbsolutePaths
   , defaultHaddockFlags
   , flagToMaybe
   , fromFlag
@@ -181,6 +183,8 @@ import Distribution.PackageDescription
   , buildable
   )
 
+import Distribution.Client.Errors
+import Distribution.Compat.ResponseFile
 import qualified Distribution.Make as Make
 import Distribution.PackageDescription.PrettyPrint
   ( writeGenericPackageDescription
@@ -202,7 +206,7 @@ import Distribution.Simple.Command
   , defaultCommandFallback
   , hiddenCommand
   )
-import Distribution.Simple.Compiler (PackageDBStack)
+import Distribution.Simple.Compiler (PackageDBStack, interpretPackageDBStack)
 import Distribution.Simple.Configure
   ( ConfigStateFileError (..)
   , configCompilerAuxEx
@@ -237,6 +241,10 @@ import Distribution.Text
   ( display
   )
 import qualified Distribution.Types.UnqualComponentName as Make
+import Distribution.Utils.Path hiding
+  ( (<.>)
+  , (</>)
+  )
 import Distribution.Verbosity as Verbosity
   ( normal
   )
@@ -248,11 +256,8 @@ import Distribution.Version
 
 import Control.Exception (AssertionFailed, assert, try)
 import Data.Monoid (Any (..))
-import Distribution.Client.Errors
-import Distribution.Compat.ResponseFile
 import System.Directory
   ( doesFileExist
-  , getCurrentDirectory
   , withCurrentDirectory
   )
 import System.Environment (getEnvironment, getExecutablePath, getProgName)
@@ -270,7 +275,7 @@ import System.IO
   , stderr
   , stdout
   )
-import System.Process (createProcess, env, proc)
+import System.Process (createProcess, env, proc, waitForProcess)
 
 -- | Entry point
 --
@@ -378,7 +383,7 @@ mainWorker args = do
       result <- try $ createProcess ((proc exec (name : cmdArgs)){env = Just new_env})
       case result of
         Left ex -> printErrors ["Error executing external command: " ++ show (ex :: SomeException)]
-        Right _ -> return ()
+        Right (_, _, _, ph) -> waitForProcess ph >>= exitWith
 
     printCommandHelp help = do
       pname <- getProgName
@@ -427,7 +432,7 @@ mainWorker args = do
       , regularCmd CmdPath.pathCommand CmdPath.pathAction
       , regularCmd genBoundsCommand genBoundsAction
       , regularCmd CmdOutdated.outdatedCommand CmdOutdated.outdatedAction
-      , wrapperCmd hscolourCommand hscolourVerbosity hscolourDistPref
+      , wrapperCmd hscolourCommand hscolourCommonFlags
       , hiddenCmd formatCommand formatAction
       , hiddenCmd actAsSetupCommand actAsSetupAction
       , hiddenCmd manpageCommand (manpageAction commandSpecs)
@@ -460,8 +465,8 @@ mainWorker args = do
           , legacyCmd testCommand testAction
           , legacyCmd benchmarkCommand benchmarkAction
           , legacyCmd cleanCommand cleanAction
-          , legacyWrapperCmd copyCommand copyVerbosity copyDistPref
-          , legacyWrapperCmd registerCommand regVerbosity regDistPref
+          , legacyWrapperCmd copyCommand copyCommonFlags
+          , legacyWrapperCmd registerCommand registerCommonFlags
           , legacyCmd reconfigureCommand reconfigureAction
           ]
 
@@ -489,35 +494,40 @@ hiddenCmd ui action =
 wrapperCmd
   :: Monoid flags
   => CommandUI flags
-  -> (flags -> Flag Verbosity)
-  -> (flags -> Flag String)
+  -> (flags -> CommonSetupFlags)
   -> CommandSpec Action
-wrapperCmd ui verbosity distPref =
-  CommandSpec ui (\ui' -> wrapperAction ui' verbosity distPref) NormalCommand
+wrapperCmd ui getCommonFlags =
+  CommandSpec ui (\ui' -> wrapperAction ui' getCommonFlags) NormalCommand
 
 wrapperAction
   :: Monoid flags
   => CommandUI flags
-  -> (flags -> Flag Verbosity)
-  -> (flags -> Flag String)
+  -> (flags -> CommonSetupFlags)
   -> Command Action
-wrapperAction command verbosityFlag distPrefFlag =
+wrapperAction command getCommonFlags =
   commandAddAction
     command
       { commandDefaultFlags = mempty
       }
     $ \flags extraArgs globalFlags -> do
-      let verbosity = fromFlagOrDefault normal (verbosityFlag flags)
+      let common = getCommonFlags flags
+          verbosity = fromFlagOrDefault normal $ setupVerbosity common
+          mbWorkDir = flagToMaybe $ setupWorkingDir common
       load <- try (loadConfigOrSandboxConfig verbosity globalFlags)
       let config = either (\(SomeException _) -> mempty) id load
-      distPref <- findSavedDistPref config (distPrefFlag flags)
-      let setupScriptOptions = defaultSetupScriptOptions{useDistPref = distPref}
+      distPref <- findSavedDistPref config (setupDistPref common)
+      let setupScriptOptions =
+            defaultSetupScriptOptions
+              { useDistPref = distPref
+              , useWorkingDir = mbWorkDir
+              }
       setupWrapper
         verbosity
         setupScriptOptions
         Nothing
         command
-        (const flags)
+        getCommonFlags
+        (const (return flags))
         (const extraArgs)
 
 configureAction
@@ -525,11 +535,12 @@ configureAction
   -> [String]
   -> Action
 configureAction (configFlags, configExFlags) extraArgs globalFlags = do
-  let verbosity = fromFlagOrDefault normal (configVerbosity configFlags)
+  let common = configCommonFlags configFlags
+      verbosity = fromFlagOrDefault normal $ setupVerbosity common
   config <-
     updateInstallDirs (configUserInstall configFlags)
       <$> loadConfigOrSandboxConfig verbosity globalFlags
-  distPref <- findSavedDistPref config (configDistPref configFlags)
+  distPref <- getSymbolicPath <$> findSavedDistPref config (setupDistPref common)
   nixInstantiate verbosity distPref True globalFlags config
   nixShell verbosity distPref globalFlags config $ do
     let configFlags' = savedConfigureFlags config `mappend` configFlags
@@ -549,7 +560,7 @@ configureAction (configFlags, configExFlags) extraArgs globalFlags = do
     withRepoContext verbosity globalFlags' $ \repoContext ->
       configure
         verbosity
-        packageDBs
+        (interpretPackageDBStack Nothing packageDBs)
         repoContext
         comp
         platform
@@ -563,11 +574,12 @@ reconfigureAction
   -> [String]
   -> Action
 reconfigureAction flags@(configFlags, _) _ globalFlags = do
-  let verbosity = fromFlagOrDefault normal (configVerbosity configFlags)
+  let common = configCommonFlags configFlags
+      verbosity = fromFlagOrDefault normal (setupVerbosity common)
   config <-
     updateInstallDirs (configUserInstall configFlags)
       <$> loadConfigOrSandboxConfig verbosity globalFlags
-  distPref <- findSavedDistPref config (configDistPref configFlags)
+  distPref <- findSavedDistPref config (setupDistPref common)
   let checkFlags = Check $ \_ saved -> do
         let flags' = saved <> flags
         unless (saved == flags') $ info verbosity message
@@ -579,7 +591,7 @@ reconfigureAction flags@(configFlags, _) _ globalFlags = do
           message =
             "flags changed: "
               ++ unwords (commandShowOptions configureExCommand flags)
-  nixInstantiate verbosity distPref True globalFlags config
+  nixInstantiate verbosity (getSymbolicPath distPref) True globalFlags config
   _ <-
     reconfigure
       configureAction
@@ -594,9 +606,10 @@ reconfigureAction flags@(configFlags, _) _ globalFlags = do
 
 buildAction :: BuildFlags -> [String] -> Action
 buildAction buildFlags extraArgs globalFlags = do
-  let verbosity = fromFlagOrDefault normal (buildVerbosity buildFlags)
+  let common = buildCommonFlags buildFlags
+      verbosity = fromFlagOrDefault normal $ setupVerbosity common
   config <- loadConfigOrSandboxConfig verbosity globalFlags
-  distPref <- findSavedDistPref config (buildDistPref buildFlags)
+  distPref <- findSavedDistPref config (setupDistPref common)
   -- Calls 'configureAction' to do the real work, so nothing special has to be
   -- done to support sandboxes.
   config' <-
@@ -609,36 +622,49 @@ buildAction buildFlags extraArgs globalFlags = do
       []
       globalFlags
       config
-  nixShell verbosity distPref globalFlags config $ do
+  nixShell verbosity (getSymbolicPath distPref) globalFlags config $ do
     build verbosity config' distPref buildFlags extraArgs
 
 -- | Actually do the work of building the package. This is separate from
 -- 'buildAction' so that 'testAction' and 'benchmarkAction' do not invoke
 -- 'reconfigure' twice.
-build :: Verbosity -> SavedConfig -> FilePath -> BuildFlags -> [String] -> IO ()
+build :: Verbosity -> SavedConfig -> SymbolicPath Pkg (Dir Dist) -> BuildFlags -> [String] -> IO ()
 build verbosity config distPref buildFlags extraArgs =
   setupWrapper
     verbosity
     setupOptions
     Nothing
     (Cabal.buildCommand progDb)
-    mkBuildFlags
+    buildCommonFlags
+    (return . mkBuildFlags)
     (const extraArgs)
   where
     progDb = defaultProgramDb
     setupOptions = defaultSetupScriptOptions{useDistPref = distPref}
 
     mkBuildFlags version = filterBuildFlags version config buildFlags'
+    commonFlags = buildCommonFlags buildFlags
     buildFlags' =
       buildFlags
-        { buildVerbosity = toFlag verbosity
-        , buildDistPref = toFlag distPref
+        { buildCommonFlags =
+            commonFlags
+              { setupVerbosity = toFlag verbosity
+              , setupDistPref = toFlag distPref
+              }
         }
 
 -- | Make sure that we don't pass new flags to setup scripts compiled against
 -- old versions of Cabal.
 filterBuildFlags :: Version -> SavedConfig -> BuildFlags -> BuildFlags
-filterBuildFlags version config buildFlags
+filterBuildFlags version config buildFlags =
+  let flags' = filterBuildFlags' version config buildFlags
+   in flags'
+        { buildCommonFlags =
+            filterCommonFlags (buildCommonFlags flags') version
+        }
+
+filterBuildFlags' :: Version -> SavedConfig -> BuildFlags -> BuildFlags
+filterBuildFlags' version config buildFlags
   | version >= mkVersion [1, 19, 1] = buildFlags_latest
   -- Cabal < 1.19.1 doesn't support 'build -j'.
   | otherwise = buildFlags_pre_1_19_1
@@ -659,11 +685,11 @@ filterBuildFlags version config buildFlags
 
 replAction :: ReplFlags -> [String] -> Action
 replAction replFlags extraArgs globalFlags = do
-  let verbosity = fromFlagOrDefault normal (replVerbosity replFlags)
+  let common = replCommonFlags replFlags
+      verbosity = fromFlagOrDefault normal $ setupVerbosity common
   config <- loadConfigOrSandboxConfig verbosity globalFlags
-  distPref <- findSavedDistPref config (replDistPref replFlags)
-  cwd <- getCurrentDirectory
-  pkgDesc <- findPackageDesc cwd
+  distPref <- findSavedDistPref config (setupDistPref common)
+  pkgDesc <- findPackageDesc Nothing
   let
     -- There is a .cabal file in the current directory: start a REPL and load
     -- the project's modules.
@@ -686,14 +712,25 @@ replAction replFlags extraArgs globalFlags = do
               { useCabalVersion = orLaterVersion $ mkVersion [1, 18, 0]
               , useDistPref = distPref
               }
+          commonFlags = replCommonFlags replFlags
           replFlags' =
             replFlags
-              { replVerbosity = toFlag verbosity
-              , replDistPref = toFlag distPref
+              { replCommonFlags =
+                  commonFlags
+                    { setupVerbosity = toFlag verbosity
+                    , setupDistPref = toFlag distPref
+                    }
               }
 
-      nixShell verbosity distPref globalFlags config $
-        setupWrapper verbosity setupOptions Nothing (Cabal.replCommand progDb) (const replFlags') (const extraArgs)
+      nixShell verbosity (getSymbolicPath distPref) globalFlags config $
+        setupWrapper
+          verbosity
+          setupOptions
+          Nothing
+          (Cabal.replCommand progDb)
+          Cabal.replCommonFlags
+          (const (return replFlags'))
+          (const extraArgs)
 
     -- No .cabal file in the current directory: just start the REPL (possibly
     -- using the sandbox package DB).
@@ -706,7 +743,7 @@ replAction replFlags extraArgs globalFlags = do
           (replProgramPaths replFlags)
           (replProgramArgs replFlags)
           programDb
-      nixShell verbosity distPref globalFlags config $ do
+      nixShell verbosity (getSymbolicPath distPref) globalFlags config $ do
         startInterpreter
           verbosity
           programDb'
@@ -728,16 +765,18 @@ installAction
   -> Action
 installAction (configFlags, _, installFlags, _, _, _) _ globalFlags
   | fromFlagOrDefault False (installOnly installFlags) = do
-      let verb = fromFlagOrDefault normal (configVerbosity configFlags)
+      let common = configCommonFlags configFlags
+          verb = fromFlagOrDefault normal (setupVerbosity common)
       config <- loadConfigOrSandboxConfig verb globalFlags
-      dist <- findSavedDistPref config (configDistPref configFlags)
+      dist <- findSavedDistPref config (setupDistPref common)
       let setupOpts = defaultSetupScriptOptions{useDistPref = dist}
       setupWrapper
         verb
         setupOpts
         Nothing
         installCommand
-        (const (mempty, mempty, mempty, mempty, mempty, mempty))
+        (const common)
+        (const (return (mempty, mempty, mempty, mempty, mempty, mempty)))
         (const [])
 installAction
   ( configFlags
@@ -749,12 +788,13 @@ installAction
     )
   extraArgs
   globalFlags = do
-    let verb = fromFlagOrDefault normal (configVerbosity configFlags)
+    let common = configCommonFlags configFlags
+        verb = fromFlagOrDefault normal $ setupVerbosity common
     config <-
       updateInstallDirs (configUserInstall configFlags)
         <$> loadConfigOrSandboxConfig verb globalFlags
 
-    dist <- findSavedDistPref config (configDistPref configFlags)
+    dist <- findSavedDistPref config $ setupDistPref common
 
     do
       targets <- readUserTargets verb extraArgs
@@ -762,7 +802,12 @@ installAction
       let configFlags' =
             maybeForceTests installFlags' $
               savedConfigureFlags config
-                `mappend` configFlags{configDistPref = toFlag dist}
+                `mappend` configFlags
+                  { configCommonFlags =
+                      (configCommonFlags configFlags)
+                        { setupDistPref = toFlag dist
+                        }
+                  }
           configExFlags' =
             defaultConfigExFlags
               `mappend` savedConfigureExFlags config
@@ -774,15 +819,30 @@ installAction
           haddockFlags' =
             defaultHaddockFlags
               `mappend` savedHaddockFlags config
-              `mappend` haddockFlags{haddockDistPref = toFlag dist}
+              `mappend` haddockFlags
+                { haddockCommonFlags =
+                    (haddockCommonFlags haddockFlags)
+                      { setupDistPref = toFlag dist
+                      }
+                }
           testFlags' =
             Cabal.defaultTestFlags
               `mappend` savedTestFlags config
-              `mappend` testFlags{testDistPref = toFlag dist}
+              `mappend` testFlags
+                { testCommonFlags =
+                    (testCommonFlags testFlags)
+                      { setupDistPref = toFlag dist
+                      }
+                }
           benchmarkFlags' =
             Cabal.defaultBenchmarkFlags
               `mappend` savedBenchmarkFlags config
-              `mappend` benchmarkFlags{benchmarkDistPref = toFlag dist}
+              `mappend` benchmarkFlags
+                { benchmarkCommonFlags =
+                    (benchmarkCommonFlags benchmarkFlags)
+                      { setupDistPref = toFlag dist
+                      }
+                }
           globalFlags' = savedGlobalFlags config `mappend` globalFlags
       (comp, platform, progdb) <- configCompilerAux' configFlags'
 
@@ -790,18 +850,16 @@ installAction
       -- future.
       progdb' <- configureAllKnownPrograms verb progdb
 
-      configFlags'' <- configAbsolutePaths configFlags'
-
       withRepoContext verb globalFlags' $ \repoContext ->
         install
           verb
-          (configPackageDB' configFlags'')
+          (interpretPackageDBStack Nothing (configPackageDB' configFlags'))
           repoContext
           comp
           platform
           progdb'
           globalFlags'
-          configFlags''
+          configFlags'
           configExFlags'
           installFlags'
           haddockFlags'
@@ -821,12 +879,15 @@ testAction
   -> GlobalFlags
   -> IO ()
 testAction (buildFlags, testFlags) extraArgs globalFlags = do
-  let verbosity = fromFlagOrDefault normal (buildVerbosity buildFlags)
+  let verbosity = fromFlagOrDefault normal (setupVerbosity $ buildCommonFlags buildFlags)
   config <- loadConfigOrSandboxConfig verbosity globalFlags
-  distPref <- findSavedDistPref config (testDistPref testFlags)
+  distPref <- findSavedDistPref config (setupDistPref $ testCommonFlags testFlags)
   let buildFlags' =
         buildFlags
-          { buildVerbosity = testVerbosity testFlags
+          { buildCommonFlags =
+              (buildCommonFlags buildFlags)
+                { setupVerbosity = setupVerbosity $ testCommonFlags testFlags
+                }
           }
       checkFlags = Check $ \_ flags@(configFlags, configExFlags) ->
         if fromFlagOrDefault False (configTests configFlags)
@@ -849,14 +910,20 @@ testAction (buildFlags, testFlags) extraArgs globalFlags = do
       []
       globalFlags
       config
-  nixShell verbosity distPref globalFlags config $ do
+  nixShell verbosity (getSymbolicPath distPref) globalFlags config $ do
     let setupOptions = defaultSetupScriptOptions{useDistPref = distPref}
-        testFlags' = testFlags{testDistPref = toFlag distPref}
+        testFlags' =
+          testFlags
+            { testCommonFlags =
+                (testCommonFlags testFlags){setupDistPref = toFlag distPref}
+            }
+        mbWorkDir = flagToMaybe $ testWorkingDir testFlags
 
     -- The package was just configured, so the LBI must be available.
     names <-
       componentNamesFromLBI
         verbosity
+        mbWorkDir
         distPref
         "test suites"
         (\c -> case c of LBI.CTest{} -> True; _ -> False)
@@ -870,7 +937,14 @@ testAction (buildFlags, testFlags) extraArgs globalFlags = do
           | otherwise = extraArgs
 
     build verbosity config distPref buildFlags' extraArgs'
-    setupWrapper verbosity setupOptions Nothing Cabal.testCommand (const testFlags') (const extraArgs')
+    setupWrapper
+      verbosity
+      setupOptions
+      Nothing
+      Cabal.testCommand
+      Cabal.testCommonFlags
+      (const (return testFlags'))
+      (const extraArgs')
 
 data ComponentNames
   = ComponentNamesUnknown
@@ -879,12 +953,14 @@ data ComponentNames
 -- | Return the names of all buildable components matching a given predicate.
 componentNamesFromLBI
   :: Verbosity
-  -> FilePath
+  -> Maybe (SymbolicPath CWD (Dir Pkg))
+  -- ^ working directory
+  -> SymbolicPath Pkg (Dir Dist)
   -> String
   -> (LBI.Component -> Bool)
   -> IO ComponentNames
-componentNamesFromLBI verbosity distPref targetsDescr compPred = do
-  eLBI <- tryGetPersistBuildConfig distPref
+componentNamesFromLBI verbosity mbWorkDir distPref targetsDescr compPred = do
+  eLBI <- tryGetPersistBuildConfig mbWorkDir distPref
   case eLBI of
     Left err -> case err of
       -- Note: the build config could have been generated by a custom setup
@@ -920,13 +996,16 @@ benchmarkAction
     let verbosity =
           fromFlagOrDefault
             normal
-            (buildVerbosity buildFlags)
+            (setupVerbosity $ buildCommonFlags buildFlags)
 
     config <- loadConfigOrSandboxConfig verbosity globalFlags
-    distPref <- findSavedDistPref config (benchmarkDistPref benchmarkFlags)
+    distPref <- findSavedDistPref config (setupDistPref $ benchmarkCommonFlags benchmarkFlags)
     let buildFlags' =
           buildFlags
-            { buildVerbosity = benchmarkVerbosity benchmarkFlags
+            { buildCommonFlags =
+                (buildCommonFlags buildFlags)
+                  { setupVerbosity = setupVerbosity $ benchmarkCommonFlags benchmarkFlags
+                  }
             }
 
     let checkFlags = Check $ \_ flags@(configFlags, configExFlags) ->
@@ -950,14 +1029,22 @@ benchmarkAction
         []
         globalFlags
         config
-    nixShell verbosity distPref globalFlags config $ do
+    nixShell verbosity (getSymbolicPath distPref) globalFlags config $ do
       let setupOptions = defaultSetupScriptOptions{useDistPref = distPref}
-          benchmarkFlags' = benchmarkFlags{benchmarkDistPref = toFlag distPref}
+          benchmarkFlags' =
+            benchmarkFlags
+              { benchmarkCommonFlags =
+                  (benchmarkCommonFlags benchmarkFlags)
+                    { setupDistPref = toFlag distPref
+                    }
+              }
+          mbWorkDir = flagToMaybe $ benchmarkWorkingDir benchmarkFlags
 
       -- The package was just configured, so the LBI must be available.
       names <-
         componentNamesFromLBI
           verbosity
+          mbWorkDir
           distPref
           "benchmarks"
           (\c -> case c of LBI.CBench{} -> True; _ -> False)
@@ -971,13 +1058,21 @@ benchmarkAction
             | otherwise = extraArgs
 
       build verbosity config' distPref buildFlags' extraArgs'
-      setupWrapper verbosity setupOptions Nothing Cabal.benchmarkCommand (const benchmarkFlags') (const extraArgs')
+      setupWrapper
+        verbosity
+        setupOptions
+        Nothing
+        Cabal.benchmarkCommand
+        Cabal.benchmarkCommonFlags
+        (const (return benchmarkFlags'))
+        (const extraArgs')
 
 haddockAction :: HaddockFlags -> [String] -> Action
 haddockAction haddockFlags extraArgs globalFlags = do
-  let verbosity = fromFlag (haddockVerbosity haddockFlags)
+  let common = haddockCommonFlags haddockFlags
+      verbosity = fromFlag $ setupVerbosity common
   config <- loadConfigOrSandboxConfig verbosity globalFlags
-  distPref <- findSavedDistPref config (haddockDistPref haddockFlags)
+  distPref <- findSavedDistPref config (setupDistPref common)
   config' <-
     reconfigure
       configureAction
@@ -988,11 +1083,17 @@ haddockAction haddockFlags extraArgs globalFlags = do
       []
       globalFlags
       config
-  nixShell verbosity distPref globalFlags config $ do
+  let mbWorkDir = flagToMaybe $ setupWorkingDir common
+  nixShell verbosity (getSymbolicPath distPref) globalFlags config $ do
     let haddockFlags' =
           defaultHaddockFlags
             `mappend` savedHaddockFlags config'
-            `mappend` haddockFlags{haddockDistPref = toFlag distPref}
+            `mappend` haddockFlags
+              { haddockCommonFlags =
+                  (haddockCommonFlags haddockFlags)
+                    { setupDistPref = toFlag distPref
+                    }
+              }
         setupScriptOptions =
           defaultSetupScriptOptions
             { useDistPref = distPref
@@ -1002,36 +1103,44 @@ haddockAction haddockFlags extraArgs globalFlags = do
       setupScriptOptions
       Nothing
       haddockCommand
-      (const haddockFlags')
+      haddockCommonFlags
+      (const (return haddockFlags'))
       (const extraArgs)
     when (haddockForHackage haddockFlags == Flag ForHackage) $ do
-      pkg <- fmap LBI.localPkgDescr (getPersistBuildConfig distPref)
-      let dest = distPref </> name <.> "tar.gz"
+      pkg <- fmap LBI.localPkgDescr (getPersistBuildConfig mbWorkDir distPref)
+      let dest = getSymbolicPath distPref </> name <.> "tar.gz"
           name = display (packageId pkg) ++ "-docs"
-          docDir = distPref </> "doc" </> "html"
+          docDir = getSymbolicPath distPref </> "doc" </> "html"
       createTarGzFile dest docDir name
       notice verbosity $ "Documentation tarball created: " ++ dest
 
 cleanAction :: CleanFlags -> [String] -> Action
 cleanAction cleanFlags extraArgs globalFlags = do
+  let common = cleanCommonFlags cleanFlags
+      verbosity = fromFlagOrDefault normal $ setupVerbosity common
   load <- try (loadConfigOrSandboxConfig verbosity globalFlags)
   let config = either (\(SomeException _) -> mempty) id load
-  distPref <- findSavedDistPref config (cleanDistPref cleanFlags)
+  distPref <- findSavedDistPref config $ setupDistPref common
   let setupScriptOptions =
         defaultSetupScriptOptions
           { useDistPref = distPref
           , useWin32CleanHack = True
           }
-      cleanFlags' = cleanFlags{cleanDistPref = toFlag distPref}
+      cleanFlags' =
+        cleanFlags
+          { cleanCommonFlags =
+              (cleanCommonFlags cleanFlags)
+                { setupDistPref = toFlag distPref
+                }
+          }
   setupWrapper
     verbosity
     setupScriptOptions
     Nothing
     cleanCommand
-    (const cleanFlags')
+    cleanCommonFlags
+    (const (return cleanFlags'))
     (const extraArgs)
-  where
-    verbosity = fromFlagOrDefault normal (cleanVerbosity cleanFlags)
 
 listAction :: ListFlags -> [String] -> Action
 listAction listFlags extraArgs globalFlags = do
@@ -1055,7 +1164,7 @@ listAction listFlags extraArgs globalFlags = do
   withRepoContext verbosity globalFlags' $ \repoContext ->
     List.list
       verbosity
-      (configPackageDB' configFlags)
+      (interpretPackageDBStack Nothing (configPackageDB' configFlags))
       repoContext
       compProgdb
       listFlags
@@ -1078,7 +1187,7 @@ infoAction infoFlags extraArgs globalFlags = do
   withRepoContext verbosity globalFlags' $ \repoContext ->
     List.info
       verbosity
-      (configPackageDB' configFlags)
+      (interpretPackageDBStack Nothing (configPackageDB' configFlags))
       repoContext
       comp
       progdb
@@ -1097,7 +1206,7 @@ fetchAction fetchFlags extraArgs globalFlags = do
   withRepoContext verbosity globalFlags' $ \repoContext ->
     fetch
       verbosity
-      (configPackageDB' configFlags)
+      (interpretPackageDBStack Nothing (configPackageDB' configFlags))
       repoContext
       comp
       platform
@@ -1111,7 +1220,7 @@ freezeAction freezeFlags _extraArgs globalFlags = do
   let verbosity = fromFlag (freezeVerbosity freezeFlags)
   config <- loadConfigOrSandboxConfig verbosity globalFlags
   distPref <- findSavedDistPref config NoFlag
-  nixShell verbosity distPref globalFlags config $ do
+  nixShell verbosity (getSymbolicPath distPref) globalFlags config $ do
     let configFlags = savedConfigureFlags config
         globalFlags' = savedGlobalFlags config `mappend` globalFlags
     (comp, platform, progdb) <- configCompilerAux' configFlags
@@ -1119,7 +1228,7 @@ freezeAction freezeFlags _extraArgs globalFlags = do
     withRepoContext verbosity globalFlags' $ \repoContext ->
       freeze
         verbosity
-        (configPackageDB' configFlags)
+        (interpretPackageDBStack Nothing (configPackageDB' configFlags))
         repoContext
         comp
         platform
@@ -1132,7 +1241,7 @@ genBoundsAction freezeFlags _extraArgs globalFlags = do
   let verbosity = fromFlag (freezeVerbosity freezeFlags)
   config <- loadConfigOrSandboxConfig verbosity globalFlags
   distPref <- findSavedDistPref config NoFlag
-  nixShell verbosity distPref globalFlags config $ do
+  nixShell verbosity (getSymbolicPath distPref) globalFlags config $ do
     let configFlags = savedConfigureFlags config
         globalFlags' = savedGlobalFlags config `mappend` globalFlags
     (comp, platform, progdb) <- configCompilerAux' configFlags
@@ -1140,7 +1249,7 @@ genBoundsAction freezeFlags _extraArgs globalFlags = do
     withRepoContext verbosity globalFlags' $ \repoContext ->
       genBounds
         verbosity
-        (configPackageDB' configFlags)
+        (interpretPackageDBStack Nothing (configPackageDB' configFlags))
         repoContext
         comp
         platform
@@ -1217,8 +1326,9 @@ uploadAction uploadFlags extraArgs globalFlags = do
         []
         globalFlags
       distPref <- findSavedDistPref config NoFlag
-      pkg <- fmap LBI.localPkgDescr (getPersistBuildConfig distPref)
-      return $ distPref </> display (packageId pkg) ++ "-docs" <.> "tar.gz"
+      let mbWorkDir = flagToMaybe $ configWorkingDir $ savedConfigureFlags config
+      pkg <- fmap LBI.localPkgDescr (getPersistBuildConfig mbWorkDir distPref)
+      return $ getSymbolicPath distPref </> display (packageId pkg) ++ "-docs" <.> "tar.gz"
 
 checkAction :: CheckFlags -> [String] -> Action
 checkAction checkFlags extraArgs _globalFlags = do
@@ -1234,13 +1344,11 @@ formatAction :: Flag Verbosity -> [String] -> Action
 formatAction verbosityFlag extraArgs _globalFlags = do
   let verbosity = fromFlag verbosityFlag
   path <- case extraArgs of
-    [] -> do
-      cwd <- getCurrentDirectory
-      tryFindPackageDesc verbosity cwd
-    (p : _) -> return p
-  pkgDesc <- readGenericPackageDescription verbosity path
+    [] -> relativeSymbolicPath <$> tryFindPackageDesc verbosity Nothing
+    (p : _) -> return $ makeSymbolicPath p
+  pkgDesc <- readGenericPackageDescription verbosity Nothing path
   -- Uses 'writeFileAtomic' under the hood.
-  writeGenericPackageDescription path pkgDesc
+  writeGenericPackageDescription (getSymbolicPath path) pkgDesc
 
 reportAction :: ReportFlags -> [String] -> Action
 reportAction reportFlags extraArgs globalFlags = do
@@ -1262,9 +1370,10 @@ reportAction reportFlags extraArgs globalFlags = do
 
 runAction :: BuildFlags -> [String] -> Action
 runAction buildFlags extraArgs globalFlags = do
-  let verbosity = fromFlagOrDefault normal (buildVerbosity buildFlags)
+  let common = buildCommonFlags buildFlags
+      verbosity = fromFlagOrDefault normal $ setupVerbosity common
   config <- loadConfigOrSandboxConfig verbosity globalFlags
-  distPref <- findSavedDistPref config (buildDistPref buildFlags)
+  distPref <- findSavedDistPref config $ setupDistPref common
   config' <-
     reconfigure
       configureAction
@@ -1275,8 +1384,9 @@ runAction buildFlags extraArgs globalFlags = do
       []
       globalFlags
       config
-  nixShell verbosity distPref globalFlags config $ do
-    lbi <- getPersistBuildConfig distPref
+  let mbWorkDir = flagToMaybe $ setupWorkingDir common
+  nixShell verbosity (getSymbolicPath distPref) globalFlags config $ do
+    lbi <- getPersistBuildConfig mbWorkDir distPref
     (exe, exeArgs) <- splitRunArgs verbosity lbi extraArgs
 
     build verbosity config' distPref buildFlags ["exe:" ++ display (exeName exe)]
@@ -1323,7 +1433,7 @@ initAction initFlags extraArgs globalFlags = do
       withRepoContext verbosity globalFlags' $ \repoContext ->
         initCmd
           verbosity
-          (configPackageDB' confFlags')
+          (interpretPackageDBStack Nothing (configPackageDB' confFlags'))
           repoContext
           comp
           progdb
@@ -1360,10 +1470,11 @@ actAsSetupAction actAsSetupFlags args _globalFlags =
    in case bt of
         Simple -> Simple.defaultMainArgs args
         Configure ->
-          Simple.defaultMainWithHooksArgs
-            Simple.autoconfUserHooks
+          Simple.defaultMainWithSetupHooksArgs
+            Simple.autoconfSetupHooks
             args
         Make -> Make.defaultMainArgs args
+        Hooks -> error "actAsSetupAction Hooks"
         Custom -> error "actAsSetupAction Custom"
 
 manpageAction :: [CommandSpec action] -> ManpageFlags -> [String] -> Action
