@@ -11,6 +11,7 @@
 #include "HsFFI.h"
 
 #include "sm/Storage.h"
+#include "linker/MMap.h"
 #include "RtsFlags.h"
 #include "RtsUtils.h"
 #include "Prelude.h"
@@ -41,6 +42,7 @@
 #include "LibdwPool.h"
 #include "sm/CNF.h"
 #include "TopHandler.h"
+#include "CheckVectorSupport.h"
 
 #if defined(PROFILING)
 # include "ProfHeap.h"
@@ -68,7 +70,7 @@
 #endif
 
 // Count of how many outstanding hs_init()s there have been.
-static int hs_init_count = 0;
+static StgWord hs_init_count = 0;
 static bool rts_shutdown = false;
 
 #if defined(mingw32_HOST_OS)
@@ -88,7 +90,7 @@ static void flushStdHandles(void);
 static void
 x86_init_fpu ( void )
 {
-#if defined(mingw32_HOST_OS) && !X86_INIT_FPU
+#if defined(mingw32_HOST_OS) && defined(x86_64_HOST_ARCH) && !X86_INIT_FPU
     /* Mingw-w64 does a stupid thing. They set the FPU precision to extended mode by default.
     The reasoning is that it's for compatibility with GNU Linux ported libraries. However the
     problem is this is incompatible with the standard Windows double precision mode.  In fact,
@@ -211,7 +213,6 @@ static void initBuiltinGcRoots(void)
     getStablePtr((StgPtr)interruptIOManager_closure);
     getStablePtr((StgPtr)ioManagerCapabilitiesChanged_closure);
 #if !defined(mingw32_HOST_OS)
-    getStablePtr((StgPtr)blockedOnBadFD_closure);
     getStablePtr((StgPtr)runHandlersPtr_closure);
 #else
     getStablePtr((StgPtr)processRemoteCompletion_closure);
@@ -242,8 +243,9 @@ hs_init_with_rtsopts(int *argc, char **argv[])
 void
 hs_init_ghc(int *argc, char **argv[], RtsConfig rts_config)
 {
-    hs_init_count++;
-    if (hs_init_count > 1) {
+    // N.B. atomic_inc returns the new value.
+    StgWord init_count = atomic_inc(&hs_init_count, 1);
+    if (init_count > 1) {
         // second and subsequent inits are ignored
         return;
     }
@@ -330,6 +332,12 @@ hs_init_ghc(int *argc, char **argv[], RtsConfig rts_config)
 #endif /* DEBUG */
     }
 
+    /* Based on the RTS flags, decide which I/O manager to use. */
+    selectIOManager();
+
+    /* Set the supported level of vector registers */
+    setVectorSupport();
+
     /* Initialize console Codepage.  */
 #if defined(mingw32_HOST_OS)
    if (is_io_mng_native_p())
@@ -339,6 +347,9 @@ hs_init_ghc(int *argc, char **argv[], RtsConfig rts_config)
     /* Initialise the adjustors subsystem */
     initAdjustors();
 
+    /* Initialise mmapForLinker */
+    initLinkerMMap();
+
     /* Initialise the stats department, phase 1 */
     initStats1();
 
@@ -452,15 +463,17 @@ hs_exit_(bool wait_foreign)
 {
     uint32_t g, i;
 
-    if (hs_init_count <= 0) {
-        errorBelch("warning: too many hs_exit()s");
-        return;
-    }
-    hs_init_count--;
-    if (hs_init_count > 0) {
+    // N.B. atomic_dec returns the new value.
+    StgInt init_count = (StgInt)atomic_dec(&hs_init_count, 1);
+    if (init_count > 0) {
         // ignore until it's the last one
         return;
     }
+    if (init_count < 0) {
+        errorBelch("warning: too many hs_exit()s");
+        return;
+    }
+
     rts_shutdown = true;
 
     /* start timing the shutdown */
@@ -488,6 +501,7 @@ hs_exit_(bool wait_foreign)
     for (g = 0; g < RtsFlags.GcFlags.generations; g++) {
         runAllCFinalizers(generations[g].weak_ptr_list);
     }
+    runAllCFinalizers(nonmoving_weak_ptr_list);
 
 #if defined(RTS_USER_SIGNALS)
     if (RtsFlags.MiscFlags.install_signal_handlers) {
