@@ -8,8 +8,10 @@
 
 #include "rts/PosixSource.h"
 #include "Rts.h"
+#include "RtsFlags.h"
 
 #include "Capability.h"
+#include "CheckVectorSupport.h"
 #include "Updates.h"
 #include "Threads.h"
 #include "STM.h"
@@ -89,13 +91,13 @@ createThread(Capability *cap, W_ size)
     stack->marking      = 0;
 
     tso = (StgTSO *)allocate(cap, sizeofW(StgTSO));
-    TICK_ALLOC_TSO();
+    TICK_ALLOC_TSO(sizeofW(StgTSO));
     SET_HDR(tso, &stg_TSO_info, CCS_SYSTEM);
 
     // Always start with the compiled code evaluator
     tso->what_next = ThreadRunGHC;
-    tso->why_blocked  = NotBlocked;
     tso->block_info.closure = (StgClosure *)END_TSO_QUEUE;
+    tso->why_blocked  = NotBlocked;
     tso->blocked_exceptions = END_BLOCKED_EXCEPTIONS_QUEUE;
     tso->bq = (StgBlockingQueue *)END_TSO_QUEUE;
     tso->flags = 0;
@@ -272,20 +274,21 @@ tryWakeupThread (Capability *cap, StgTSO *tso)
     traceEventThreadWakeup (cap, tso, tso->cap->no);
 
 #if defined(THREADED_RTS)
-    if (tso->cap != cap)
+    Capability *tso_owner = RELAXED_LOAD(&tso->cap);
+    if (tso_owner != cap)
     {
         MessageWakeup *msg;
         msg = (MessageWakeup *)allocate(cap,sizeofW(MessageWakeup));
         msg->tso = tso;
         SET_HDR(msg, &stg_MSG_TRY_WAKEUP_info, CCS_SYSTEM);
-        sendMessage(cap, tso->cap, (Message*)msg);
+        sendMessage(cap, tso_owner, (Message*)msg);
         debugTraceCap(DEBUG_sched, cap, "message: try wakeup thread %"
-                      FMT_StgThreadID " on cap %d", tso->id, tso->cap->no);
+                      FMT_StgThreadID " on cap %d", tso->id, tso_owner->no);
         return;
     }
 #endif
 
-    switch (tso->why_blocked)
+    switch (ACQUIRE_LOAD(&tso->why_blocked))
     {
     case BlockedOnMVar:
     case BlockedOnMVarRead:
@@ -437,7 +440,7 @@ checkBlockingQueues (Capability *cap, StgTSO *tso)
         p = UNTAG_CLOSURE(bq->bh);
         const StgInfoTable *pinfo = ACQUIRE_LOAD(&p->header.info);
         if (pinfo != &stg_BLACKHOLE_info ||
-            ((StgInd *)p)->indirectee != (StgClosure*)bq)
+            (RELAXED_LOAD(&((StgInd *)p)->indirectee) != (StgClosure*)bq))
         {
             wakeBlockingQueue(cap,bq);
         }
@@ -468,7 +471,7 @@ updateThunk (Capability *cap, StgTSO *tso, StgClosure *thunk, StgClosure *val)
         return;
     }
 
-    v = UNTAG_CLOSURE(((StgInd*)thunk)->indirectee);
+    v = UNTAG_CLOSURE(ACQUIRE_LOAD(&((StgInd*)thunk)->indirectee));
 
     updateWithIndirection(cap, thunk, val);
 
@@ -686,9 +689,25 @@ threadStackOverflow (Capability *cap, StgTSO *tso)
             // overflow, being replaced by a non-moving 32k chunk.
             //
         } else {
+
             new_stack->sp -= sizeofW(StgUnderflowFrame);
             frame = (StgUnderflowFrame*)new_stack->sp;
-            frame->info = &stg_stack_underflow_frame_info;
+
+            // See Note [realArgRegsCover] in GHC.Cmm.CallConv.
+            switch (vectorSupportGlobalVar) {
+              case 3:
+                frame->info = &stg_stack_underflow_frame_v64_info;
+                break;
+              case 2:
+                frame->info = &stg_stack_underflow_frame_v32_info;
+                break;
+              case 1:
+                frame->info = &stg_stack_underflow_frame_v16_info;
+                break;
+              default:
+                frame->info = &stg_stack_underflow_frame_d_info;
+                break;
+            }
             frame->next_chunk  = old_stack;
         }
 
@@ -734,7 +753,10 @@ threadStackUnderflow (Capability *cap, StgTSO *tso)
 
     frame = (StgUnderflowFrame*)(old_stack->stack + old_stack->stack_size
                                  - sizeofW(StgUnderflowFrame));
-    ASSERT(frame->info == &stg_stack_underflow_frame_info);
+    ASSERT( frame->info == &stg_stack_underflow_frame_d_info
+         || frame->info == &stg_stack_underflow_frame_v16_info
+         || frame->info == &stg_stack_underflow_frame_v32_info
+         || frame->info == &stg_stack_underflow_frame_v64_info );
 
     new_stack = (StgStack*)frame->next_chunk;
     tso->stackobj = new_stack;
@@ -808,7 +830,7 @@ loop:
     qinfo = ACQUIRE_LOAD(&q->header.info);
     if (qinfo == &stg_IND_info ||
         qinfo == &stg_MSG_NULL_info) {
-        q = (StgMVarTSOQueue*)((StgInd*)q)->indirectee;
+        q = (StgMVarTSOQueue*) ACQUIRE_LOAD(&((StgInd*)q)->indirectee);
         goto loop;
     }
 
@@ -825,10 +847,11 @@ loop:
         }
     }
 
-    ASSERT(tso->block_info.closure == (StgClosure*)mvar);
     // save why_blocked here, because waking up the thread destroys
     // this information
-    StgWord why_blocked = RELAXED_LOAD(&tso->why_blocked);
+    StgWord why_blocked = ACQUIRE_LOAD(&tso->why_blocked);
+    ASSERT(why_blocked == BlockedOnMVarRead || why_blocked == BlockedOnMVar);
+    ASSERT(tso->block_info.closure == (StgClosure*)mvar);
 
     // actually perform the takeMVar
     StgStack* stack = tso->stackobj;
@@ -873,7 +896,7 @@ StgMutArrPtrs *listThreads(Capability *cap)
     StgMutArrPtrs *arr =
         (StgMutArrPtrs *)allocate(cap, sizeofW(StgMutArrPtrs) + size);
     SET_HDR(arr, &stg_MUT_ARR_PTRS_DIRTY_info, CCS_SYSTEM);
-    TICK_ALLOC_PRIM(sizeofW(StgMutArrPtrs), n, 0);
+    TICK_ALLOC_PRIM(sizeofW(StgMutArrPtrs), size, 0);
     arr->ptrs = n_threads;
     arr->size = size;
 
@@ -902,7 +925,7 @@ StgMutArrPtrs *listThreads(Capability *cap)
 void
 printThreadBlockage(StgTSO *tso)
 {
-  switch (tso->why_blocked) {
+  switch (ACQUIRE_LOAD(&tso->why_blocked)) {
 #if defined(mingw32_HOST_OS)
     case BlockedOnDoProc:
     debugBelch("is blocked on proc (request: %u)", tso->block_info.async_result->reqID);
@@ -1007,6 +1030,20 @@ printAllThreads(void)
   }
 }
 
+void
+printGlobalThreads(void)
+{
+  for (uint32_t g = 0; g < RtsFlags.GcFlags.generations; g++) {
+    debugBelch("\ngen %d\n", g);
+    for (StgTSO *t = generations[g].threads; t != END_TSO_QUEUE; t = t->global_link) {
+      debugBelch("thread %p (id=%lu)\n", t, (unsigned long)t->id);
+    }
+    for (StgTSO *t = generations[g].old_threads; t != END_TSO_QUEUE; t = t->global_link) {
+      debugBelch("thread %p (id=%lu) (old)\n", t, (unsigned long)t->id);
+    }
+  }
+}
+
 // useful from gdb
 void
 printThreadQueue(StgTSO *t)
