@@ -61,6 +61,7 @@
 #include "Trace.h"
 #include "Sanity.h"
 #include "Capability.h"
+#include "IOManager.h"
 #include "LdvProfile.h"
 #include "HeapUtils.h"
 #include "Hash.h"
@@ -137,23 +138,30 @@ scavengeTSO (StgTSO *tso)
         evacuate((StgClosure **)&tso->label);
     }
 
-    if (   tso->why_blocked == BlockedOnMVar
-        || tso->why_blocked == BlockedOnMVarRead
-        || tso->why_blocked == BlockedOnBlackHole
-        || tso->why_blocked == BlockedOnMsgThrowTo
-        || tso->why_blocked == NotBlocked
-        ) {
+    switch (ACQUIRE_LOAD(&tso->why_blocked)) {
+    case BlockedOnMVar:
+    case BlockedOnMVarRead:
+    case BlockedOnBlackHole:
+    case BlockedOnMsgThrowTo:
+    case NotBlocked:
         evacuate(&tso->block_info.closure);
-    }
+        break;
+    case BlockedOnRead:
+    case BlockedOnWrite:
+    case BlockedOnDelay:
+    case BlockedOnDoProc:
+        scavengeTSOIOManager(tso);
+        break;
+    default:
 #if defined(THREADED_RTS)
     // in the THREADED_RTS, block_info.closure must always point to a
     // valid closure, because we assume this in throwTo().  In the
     // non-threaded RTS it might be a FD (for
     // BlockedOnRead/BlockedOnWrite) or a time value (BlockedOnDelay)
-    else {
         tso->block_info.closure = (StgClosure *)END_TSO_QUEUE;
-    }
 #endif
+        break;
+    }
 
     tso->dirty = gct->failed_to_evac;
 
@@ -1593,6 +1601,14 @@ scavenge_one(StgPtr p)
 #endif
       break;
 
+    case BCO: {
+        StgBCO *bco = (StgBCO *)p;
+        evacuate((StgClosure **)&bco->instrs);
+        evacuate((StgClosure **)&bco->literals);
+        evacuate((StgClosure **)&bco->ptrs);
+        break;
+    }
+
     case COMPACT_NFDATA:
         scavenge_compact((StgCompactNFData*)p);
         break;
